#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Task for Branin Function
=========================
"""
import copy
import math

import numpy

from orion.benchmark.task.base import BaseTask


class Branin(BaseTask):
    """`Branin function <http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Branin01>`_
    as benchmark task
    """

    def __init__(self, max_trials=20):
        super(Branin, self).__init__(max_trials=max_trials)

    def call(self, x):
        """Evaluate a 2-D branin function."""
        x = copy.deepcopy(x)
        x[0] = (x[0] * 15) - 5
        x[1] = x[1] * 15

        a = 1
        b = 5.1 / (4 * numpy.square(math.pi))
        c = 5 / math.pi
        r = 6
        t = 1 / (8 * math.pi)
        s = 10
        y = (
            a * numpy.square(x[1] - b * numpy.square(x[0]) + c * x[0] - r)
            + s * (1 - t) * numpy.cos(x[0])
            + s
        )

        return [dict(name="branin", type="objective", value=y)]

    def get_search_space(self):
        """Return the search space for the task objective function"""
        rspace = {"x": "uniform(0, 1, shape=2, precision=10)"}

        return rspace



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Installation script for `orion.algo.gradient_descent`."""
from setuptools import setup

setup_args = dict(
    name="orion.algo.gradient_descent",
    version=0.1,
    description="Implement a gradient descent algorithm, for demo and testing.",
    license="BSD-3-Clause",
    author="Christos Tsirigotis",
    author_email="tsirif@gmail.com",
    url="https://github.com/epistimio/orion",
    packages=["orion.algo"],
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        "OptimizationAlgorithm": [
            "gradient_descent = orion.algo.gradient_descent:Gradient_Descent"
        ],
    },
    install_requires=["orion"],
    setup_requires=["setuptools"],
    # "Zipped eggs don't play nicely with namespace packaging"
    # from https://github.com/pypa/sample-namespace-packages
    zip_safe=False,
)

setup_args["keywords"] = [
    "Machine Learning",
    "Deep Learning",
    "Distributed",
    "Optimization",
]

setup_args["platforms"] = ["Linux"]

setup_args["classifiers"] = [
    "Development Status :: 1 - Planning",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
] + [("Programming Language :: Python :: %s" % x) for x in "3 3.6 3.7 3.8".split()]

if __name__ == "__main__":
    setup(**setup_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helper functions for terminal i/o
=================================
"""


def ask_question(question, default=None, choice=None, ignore_case=False):
    """Ask a question to the user and receive an answer.

    Parameters
    ----------
    question: str
        The question to be asked.
    default: str
        The default value to use if the user enters nothing.
    choice: list
        List of expected values to check user answer
    ignore_case: bool
        Used only if choice is provided. If True, ignore case when checking
        user answer against given choice.

    Returns
    -------
    str
        The answer provided by the user.

    """
    if choice is not None:
        if ignore_case:
            choice = [value.lower() for value in choice]
        question = question + " (choice: {})".format(", ".join(choice))

    if default is not None:
        question = question + " (default: {}) ".format(default)

    while True:
        answer = input(question)
        if answer.strip() == "":
            answer = default
            break
        if choice is None:
            break
        if answer in choice or (ignore_case and answer.lower() in choice):
            break
        print(
            "Unexpected value: {}. Must be one of: {}\n".format(
                answer, ", ".join(choice)
            )
        )

    return answer


def confirm_name(message, name, force=False):
    """Ask the user to confirm the name.

    Parameters
    ----------
    message: str
        The message to be printed.
    name: str
        The string that the user must enter.
    force: bool
        Override confirmation and return True. Default: False.

    Returns
    -------
    bool
        True if confirmed, False otherwise.

    """
    if force:
        print(message)
        print("FORCED")
        return True

    answer = input(message)

    return answer.strip() == name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for algos included with orion."""
import copy
import random

import numpy
import pytest

from orion.client import create_experiment, workon
from orion.testing.state import OrionState

storage = {"type": "legacy", "database": {"type": "ephemeraldb"}}


space = {"x": "uniform(-50, 50)"}


space_with_fidelity = {"x": space["x"], "noise": "fidelity(1,10,4)"}


algorithm_configs = {
    "random": {"random": {"seed": 1}},
    "gridsearch": {"gridsearch": {"n_values": 40}},
    "tpe": {
        "tpe": {
            "seed": 1,
            "n_initial_points": 20,
            "n_ei_candidates": 24,
            "gamma": 0.25,
            "equal_weight": False,
            "prior_weight": 1.0,
            "full_weight_num": 25,
        }
    },
    "asha": {"asha": {"seed": 1, "num_rungs": 4, "num_brackets": 1, "repetitions": 2}},
    "hyperband": {"hyperband": {"repetitions": 5, "seed": 1}},
    "evolutiones": {
        "evolutiones": {
            "mutate": None,
            "repetitions": 5,
            "nums_population": 20,
            "seed": 1,
            "max_retries": 100,
        }
    },
}

no_fidelity_algorithms = ["random", "tpe", "gridsearch"]
no_fidelity_algorithm_configs = {
    key: algorithm_configs[key] for key in no_fidelity_algorithms
}

fidelity_only_algorithms = ["asha", "hyperband", "evolutiones"]
fidelity_only_algorithm_configs = {
    key: algorithm_configs[key] for key in fidelity_only_algorithms
}


def rosenbrock(x, noise=None):
    """Evaluate partial information of a quadratic."""
    z = x - 34.56789
    if noise:
        noise = (1 - noise / 10) + 0.0001
        z *= random.gauss(0, noise)

    return [
        {"name": "objective", "type": "objective", "value": 4 * z ** 2 + 23.4},
        {"name": "gradient", "type": "gradient", "value": [8 * z]},
    ]


def multidim_rosenbrock(x, noise=None, shape=(2, 1)):
    x = numpy.array(x)
    assert x.shape == shape
    return rosenbrock(x.reshape(-1)[0], noise)


@pytest.mark.parametrize(
    "algorithm",
    fidelity_only_algorithm_configs.values(),
    ids=list(fidelity_only_algorithm_configs.keys()),
)
def test_missing_fidelity(algorithm):
    """Test a simple usage scenario."""
    with pytest.raises(RuntimeError) as exc:
        workon(rosenbrock, space, algorithms=algorithm, max_trials=100)

    assert "https://orion.readthedocs.io/en/develop/user/algorithms.html" in str(
        exc.value
    )


@pytest.mark.parametrize(
    "algorithm",
    no_fidelity_algorithm_configs.values(),
    ids=list(no_fidelity_algorithm_configs.keys()),
)
def test_simple(algorithm):
    """Test a simple usage scenario."""
    max_trials = 30
    exp = workon(rosenbrock, space, algorithms=algorithm, max_trials=max_trials)

    assert exp.max_trials == max_trials
    assert exp.configuration["algorithms"] == algorithm

    trials = exp.fetch_trials()
    assert len(trials) == max_trials
    assert trials[-1].status == "completed"

    best_trial = sorted(trials, key=lambda trial: trial.objective.value)[0]
    assert best_trial.objective.name == "objective"
    assert abs(best_trial.objective.value - 23.4) < 15
    assert len(best_trial.params) == 1
    param = best_trial._params[0]
    assert param.name == "x"
    assert param.type == "real"


@pytest.mark.parametrize(
    "algorithm",
    no_fidelity_algorithm_configs.values(),
    ids=list(no_fidelity_algorithm_configs.keys()),
)
def test_cardinality_stop(algorithm):
    """Test when algo needs to stop because all space is explored (dicrete space)."""
    discrete_space = copy.deepcopy(space)
    discrete_space["x"] = "uniform(-10, 5, discrete=True)"
    exp = workon(rosenbrock, discrete_space, algorithms=algorithm, max_trials=30)

    trials = exp.fetch_trials()
    assert len(trials) == 16
    assert trials[-1].status == "completed"


@pytest.mark.parametrize(
    "algorithm", algorithm_configs.values(), ids=list(algorithm_configs.keys())
)
def test_with_fidelity(algorithm):
    """Test a scenario with fidelity."""
    exp = workon(rosenbrock, space_with_fidelity, algorithms=algorithm, max_trials=30)

    assert exp.configuration["algorithms"] == algorithm

    trials = exp.fetch_trials()
    assert len(trials) >= 30
    assert trials[29].status == "completed"

    trials = [trial for trial in trials if trial.status == "completed"]
    results = [trial.objective.value for trial in trials]
    best_trial = next(iter(sorted(trials, key=lambda trial: trial.objective.value)))

    assert best_trial.objective.name == "objective"
    assert abs(best_trial.objective.value - 23.4) < 5
    assert len(best_trial.params) == 2
    fidelity = best_trial._params[0]
    assert fidelity.name == "noise"
    assert fidelity.type == "fidelity"
    assert fidelity.value in [1, 2, 5, 10]
    param = best_trial._params[1]
    assert param.name == "x"
    assert param.type == "real"


@pytest.mark.parametrize(
    "algorithm", algorithm_configs.values(), ids=list(algorithm_configs.keys())
)
def test_with_multidim(algorithm):
    """Test a scenario with a dimension shape > 1."""
    space = copy.deepcopy(space_with_fidelity)
    space["x"] = "uniform(-50, 50, shape=(2, 1))"
    MAX_TRIALS = 30
    exp = workon(
        multidim_rosenbrock, space, algorithms=algorithm, max_trials=MAX_TRIALS
    )

    assert exp.configuration["algorithms"] == algorithm

    trials = exp.fetch_trials()
    assert len(trials) >= 25  # Grid-search is a bit smaller
    completed_trials = exp.fetch_trials_by_status("completed")
    assert len(completed_trials) >= MAX_TRIALS or len(completed_trials) == len(trials)

    trials = completed_trials
    results = [trial.objective.value for trial in trials]
    best_trial = next(iter(sorted(trials, key=lambda trial: trial.objective.value)))

    assert best_trial.objective.name == "objective"
    assert abs(best_trial.objective.value - 23.4) < 5
    assert len(best_trial.params) == 2
    fidelity = best_trial._params[0]
    assert fidelity.name == "noise"
    assert fidelity.type == "fidelity"
    assert fidelity.value in [1, 2, 5, 10]
    param = best_trial._params[1]
    assert param.name == "x"
    assert param.type == "real"


@pytest.mark.parametrize(
    "algorithm", algorithm_configs.values(), ids=list(algorithm_configs.keys())
)
def test_with_evc(algorithm):
    """Test a scenario where algos are warm-started with EVC."""

    with OrionState(storage={"type": "legacy", "database": {"type": "PickledDB"}}):
        base_exp = create_experiment(
            name="exp",
            space=space_with_fidelity,
            algorithms=algorithm_configs["random"],
            max_trials=10,
        )
        base_exp.workon(rosenbrock, max_trials=10)

        exp = create_experiment(
            name="exp",
            space=space_with_fidelity,
            algorithms=algorithm,
            max_trials=30,
            branching={"branch_from": "exp"},
        )

        assert exp.version == 2

        exp.workon(rosenbrock, max_trials=30)

        assert exp.configuration["algorithms"] == algorithm

        trials = exp.fetch_trials(with_evc_tree=False)

        # Some algo may not be able to suggest exactly 30 trials (ex: hyperband)
        assert len(trials) >= 20

        trials_with_evc = exp.fetch_trials(with_evc_tree=True)
        assert len(trials_with_evc) >= 30
        assert len(trials_with_evc) - len(trials) == 10

        completed_trials = [
            trial for trial in trials_with_evc if trial.status == "completed"
        ]
        assert len(completed_trials) == 30

        results = [trial.objective.value for trial in completed_trials]
        best_trial = next(
            iter(sorted(completed_trials, key=lambda trial: trial.objective.value))
        )

        assert best_trial.objective.name == "objective"
        assert abs(best_trial.objective.value - 23.4) < 1e-5
        assert len(best_trial.params) == 2
        fidelity = best_trial._params[0]
        assert fidelity.name == "noise"
        assert fidelity.type == "fidelity"
        assert fidelity.value == 10
        param = best_trial._params[1]
        assert param.name == "x"
        assert param.type == "real"


@pytest.mark.parametrize(
    "algorithm", algorithm_configs.values(), ids=list(algorithm_configs.keys())
)
def test_parallel_workers(algorithm):
    """Test parallel execution with joblib"""
    with OrionState() as cfg:  # Using PickledDB

        name = "{}_exp".format(list(algorithm.keys())[0])

        base_exp = create_experiment(
            name=name,
            space=space_with_fidelity,
            algorithms=algorithm_configs["random"],
        )
        base_exp.workon(rosenbrock, max_trials=10)

        exp = create_experiment(
            name=name,
            space=space_with_fidelity,
            algorithms=algorithm,
            branching={"branch_from": name},
        )

        assert exp.version == 2

        exp.workon(rosenbrock, max_trials=30, n_workers=2)

        assert exp.configuration["algorithms"] == algorithm

        trials = exp.fetch_trials(with_evc_tree=False)
        assert len(trials) >= 20

        trials_with_evc = exp.fetch_trials(with_evc_tree=True)
        assert len(trials_with_evc) >= 30
        assert len(trials_with_evc) - len(trials) == 10

        completed_trials = [
            trial for trial in trials_with_evc if trial.status == "completed"
        ]
        assert 30 <= len(completed_trials) <= 30 + 2

        results = [trial.objective.value for trial in completed_trials]
        best_trial = next(
            iter(sorted(completed_trials, key=lambda trial: trial.objective.value))
        )

        assert best_trial.objective.name == "objective"
        assert abs(best_trial.objective.value - 23.4) < 1e-5
        assert len(best_trial.params) == 2
        fidelity = best_trial._params[0]
        assert fidelity.name == "noise"
        assert fidelity.type == "fidelity"
        assert fidelity.value == 10
        param = best_trial._params[1]
        assert param.name == "x"
        assert param.type == "real"



# -*- coding: utf-8 -*-
"""
Helpers for building responses according to the specification
=============================================================

Offers functions and attributes to generate response objects according to the API specification.

"""
from orion.core.worker.experiment import Experiment
from orion.core.worker.trial import Trial

ERROR_EXPERIMENT_NOT_FOUND = "Experiment not found"
ERROR_INVALID_PARAMETER = "Invalid parameter"
ERROR_TRIAL_NOT_FOUND = "Trial not found"


def build_trial_response(trial: Trial) -> dict:
    """
    Build the response representing a trial according to the API specification.

    Parameters
    ----------
    trial: Trial
        The trial to return for the API.

    Returns
    -------
    A JSON-serializable dict representing the given trial.

    """
    return {
        "id": trial.id,
        "submitTime": str(trial.submit_time),
        "startTime": str(trial.start_time),
        "endTime": str(trial.end_time),
        "parameters": trial.params,
        "objective": trial.objective.value,
        "statistics": {
            statistic.name: statistic.value for statistic in trial.statistics
        },
    }


def build_experiment_response(
    experiment: Experiment, status: str, algorithm: dict, best_trial: Trial = None
):
    """
    Build the response representing an experiment response object according to the API
    specification.

    Parameters
    ----------
    experiment: Experiment
        The experiment to return to the API
    status: str
        The status of the experiment
    algorithm: dict
        The dictionary containing the algorithm's configuration
    best_trial: Trial (Optional)
        The best trial to date of the experiment

    Returns
    -------
    A JSON-serializable experiment response object representing the given experiment.
    """
    return {
        "name": experiment.name,
        "version": experiment.version,
        "status": status,
        "trialsCompleted": experiment.stats["trials_completed"]
        if experiment.stats
        else 0,
        "startTime": str(experiment.stats["start_time"]) if experiment.stats else None,
        "endTime": str(experiment.stats["finish_time"]) if experiment.stats else None,
        "user": experiment.metadata["user"],
        "orionVersion": experiment.metadata["orion_version"],
        "config": {
            "maxTrials": experiment.max_trials,
            "maxBroken": experiment.max_broken,
            "poolSize": experiment.pool_size,
            "algorithm": algorithm,
            "space": experiment.configuration["space"],
        },
        "bestTrial": build_trial_response(best_trial) if best_trial else {},
    }


def build_experiments_response(experiments: dict):
    """
    Build the response representing a list of experiments according to the API specification.

    Parameters
    ----------
    experiments: dict
        A dict containing pairs of ``experiment-name:experiment-version``.

    Returns
    -------
    A JSON-serializable list of experiments as defined in the API specification.
    """
    result = []
    for name, version in experiments.items():
        result.append({"name": name, "version": version})
    return result


def build_trials_response(trials: list):
    """
    Build the response representing a list of trials according to the API specification.

    Parameters
    ----------
    trials: list
        A list of :class:`orion.core.worker.trial.Trial`.

    Returns
    -------
    A JSON-serializable list of trials as defined in the API specification.
    """
    response = []
    for trial in trials:
        response.append({"id": trial.id})
    return response



"""
==========================
Local Parameter Importance
==========================

.. hint:: 

   Conveys a very compact measure of the importance of the different hyperparameters
   to achieve the best objective found so far.

The local parameter importance measures the variance of the results when
varying one hyperparameter and keeping all other fixes [Biedenkapp2018]_.

Given a best set of hyperparameters, we separately build a grid
for each hyperparameter and compute the variance of the results when keeping all other
hyperparameter fixed to their values in the best set. In order to infer these results,
we train a regression model from scikit-learn (by default RandomForestRegressor)
on the trial history of the experiment, and use it to predict the objective.
The ratio of variance for one hyperparameter versus the sum of variances for all hyperparameters
is used as the local parameter importance metric.

.. autofunction:: orion.plotting.base.lpi
    :noindex:

The local parameter importance plot can be executed directly from the ``experiment`` with
``plot.lpi()`` as shown in the example below.

"""
from orion.client import get_experiment

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = dict(type="legacy", database=dict(type="pickleddb", host="../db.pkl"))

# Load the data for the specified experiment
experiment = get_experiment("2-dim-exp", storage=storage)
fig = experiment.plot.lpi()
fig

#%%
# On this plot the x-axis shows the different hyperparameters while the y-axis gives the
# local parameter importance.
# The error bars represent the standard deviation of the LPI based on 10 runs. Remember
# that the LPI requires the training of a regression model. The initial state of this model can
# greatly influence the results. This is why the computation of the LPI is done multiple times
# (10 times by default) using different random seeds. Here is an example
# setting the number of points for the grids, the number of runs and the initial random seed
# for the regression model.

experiment.plot.lpi(n_points=10, n_runs=5, model_kwargs=dict(random_state=1))

#%%
# We can see that the learning rate had a larger impact than the dropout in achieving the best
# objective.
# A search space of only 2 dimensions is easy to analyse visually however so the LPI
# has little additional value in this example. We need to use an example with a larger search space
# to better show to utility of the LPI. We will load results from tutorial
# :ref:`sphx_glr_auto_tutorials_code_2_hyperband_checkpoint.py` for this.

# Load the data for the specified experiment
experiment = get_experiment("hyperband-cifar10", storage=storage)
experiment.plot.lpi()

#%%
# There is a large difference here between the most important hyperparameter (learning rate) and the
# least important one (gamma).
#
# One caveat of LPI is that the variance for each hyperparameters depends on the search space.
# If the prior for one hyperparameter is narrow and fits the region of best values for this
# hyperparameter, then the variance will be low and this hyperparameter will be considered
# non-important. It may be important, but it is not important to optimize it within this narrow
# search space. Another related issue, is that if one hyperparameter have a dramatic effect, it will
# lead to a variance so large that the other hyperparameters will seem unrelevant in comparison.
# This is what we observe here with the learning rate. If we branch from the experiment and define
# a narrowed prior for the learning rate, we will see that it becomes an unimportant
# hyperparameter.
# See documentation on :ref:`EVC system` for more information on branching, or
# :py:func:`orion.client.build_experiment` for informations on ``branching`` arguments.
# Original learning rate prior was ``loguniform(1e-5, 0.1)``. We will narrow it to
# ``loguniform(1e-3, 0.1)``.

from orion.client import build_experiment

# Branch from "hyperband-cifar10" with a narrower search space.
experiment = build_experiment(
    "narrow-hyperband-cifar10",
    branching={"branch_from": "hyperband-cifar10"},
    space={
        "epochs": "fidelity(1, 120, base=4)",
        "learning_rate": "loguniform(1e-3, 0.1)",
        "momentum": "uniform(0, 0.9)",
        "weight_decay": "loguniform(1e-10, 1e-2)",
        "gamma": "loguniform(0.97, 1)",
    },
    storage=storage,
)

experiment.plot.lpi()

#%%
# The prior of the learning rate is arguably large, spanning over 3 orders of magnitude
# `(0.001, 0.1)`. Nevertheless, for this problem, most learning rates within this range
# leads to optimal results whenever the other hyperparameters are optimal. What you must remember
# is that defining to narrow search spaces may lead to misleading local parameter importance.
# See :ref:`sphx_glr_auto_examples_plot_4_partial_dependencies.py` for a visualization to verify if
# the search space you defined may be too narrow.
#
# Special cases
# -------------
#
# Logarithmic scale
# ~~~~~~~~~~~~~~~~~
#
# Dimensions with a logarithmic prior :ref:`search-space-prior-loguniform` are linearized before
# being passed to the regression model (using log(dim) instead of dim directly). This means the
# model is trained and will be making predictions in the linearized space.
#
# Dimension with shape
# ~~~~~~~~~~~~~~~~~~~~
#
# If some dimensions have a :ref:`search-space-shape` larger than 1, they will be flattened so that
# each subdimension can be represented in the bar plot.

# Load the data for the specified experiment
experiment = get_experiment("2-dim-shape-exp", storage=storage)
experiment.plot.lpi()

#%%
# In the example above, the dimension ``learning_rate~loguniform(1e-5, 1e-2, shape=3)``
# is flattened and represented with ``learning_rate[i]``. If the shape would be or more dimensions
# (ex: ``(3, 2)``), the indices would be ``learning_rate[i,j]`` with i=0..2 and j=0..1.

#%%
# Categorical dimension
# ~~~~~~~~~~~~~~~~~~~~~
#
# Categorical dimensions are converted into integer values, so that the regression model
# can handle them. The integers are simply indices that are assigned to each category in arbitrary
# order. Here is an example where dimension ``mt-join`` has the prior
# ``choices(['mean', 'max', 'concat'])``.

# Load the data for the specified experiment
experiment = get_experiment("3-dim-cat-shape-exp", storage=storage)
experiment.plot.lpi()

#%%
# Finally we save the image to serve as a thumbnail for this example. See
# the guide
# :ref:`How to save <sphx_glr_auto_examples_how-tos_code_2_how_to_save.py>`
# for more information on image saving.

fig.write_image("../../docs/src/_static/lpi_thumbnail.png")

# sphinx_gallery_thumbnail_path = '_static/lpi_thumbnail.png'

#%%
# .. [Biedenkapp2018] Biedenkapp, André, et al.
#    "Cave: Configuration assessment, visualization and evaluation."
#    International Conference on Learning and Intelligent Optimization.
#    Springer, Cham, 2018.



# -*- coding: utf-8 -*-
"""
Input/output functions for interaction with user's script
=========================================================

Parsers and formatters to be used with user's input files or cli arguments and expected outputs.

"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures and utils for unittests and functional tests."""
import copy
import os
import zlib

import pytest
import yaml
from pymongo import MongoClient

import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.core.io.database import Database
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage


@pytest.fixture()
def exp_config():
    """Load an example database."""
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "experiment.yaml")
    ) as f:
        exp_config = list(yaml.safe_load_all(f))

    for config in exp_config[0]:
        backward.populate_space(config)

    return exp_config


@pytest.fixture
def only_experiments_db(storage, exp_config):
    """Clean the database and insert only experiments."""
    for exp in exp_config[0]:
        storage.create_experiment(exp)


def ensure_deterministic_id(name, storage, version=1, update=None):
    """Change the id of experiment to its name."""
    experiment = storage.fetch_experiments({"name": name, "version": version})[0]
    storage.delete_experiment(uid=experiment["_id"])
    _id = zlib.adler32(str((name, version)).encode())
    experiment["_id"] = _id

    if experiment["refers"]["parent_id"] is None:
        experiment["refers"]["root_id"] = _id

    if update is not None:
        experiment.update(update)

    storage.create_experiment(experiment)


# Experiments combinations fixtures
@pytest.fixture
def one_experiment(monkeypatch, storage):
    """Create an experiment without trials."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    name = "test_single_exp"
    orion.core.cli.main(
        ["hunt", "--init-only", "-n", name, "./black_box.py", "--x~uniform(0,1)"]
    )
    ensure_deterministic_id(name, storage)
    return get_storage().fetch_experiments({"name": name})[0]


@pytest.fixture
def one_experiment_changed_vcs(one_experiment):
    """Create an experiment without trials."""
    experiment = experiment_builder.build(name=one_experiment["name"])

    experiment.metadata["VCS"] = {
        "type": "git",
        "is_dirty": False,
        "HEAD_sha": "new",
        "active_branch": "master",
        "diff_sha": None,
    }

    get_storage().update_experiment(experiment, metadata=experiment.metadata)


@pytest.fixture
def one_experiment_no_version(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    one_experiment["name"] = one_experiment["name"] + "-no-version"
    one_experiment.pop("version")

    def fetch_without_version(query, selection=None):
        if query.get("name") == one_experiment["name"] or query == {}:
            return [copy.deepcopy(one_experiment)]

        return []

    monkeypatch.setattr(get_storage(), "fetch_experiments", fetch_without_version)

    return one_experiment


@pytest.fixture
def with_experiment_using_python_api(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    experiment = experiment_builder.build(
        name="from-python-api", space={"x": "uniform(0, 10)"}
    )

    return experiment


@pytest.fixture
def with_experiment_missing_conf_file(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    exp = experiment_builder.build(name="test_single_exp", version=1)
    conf_file = "idontexist.yaml"
    exp.metadata["user_config"] = conf_file
    exp.metadata["user_args"] += ["--config", conf_file]
    Database().write("experiments", exp.configuration, query={"_id": exp.id})

    return exp


@pytest.fixture
def broken_refers(one_experiment, storage):
    """Create an experiment with broken refers."""
    ensure_deterministic_id(
        "test_single_exp", storage, update=dict(refers={"oups": "broken"})
    )


@pytest.fixture
def single_without_success(one_experiment):
    """Create an experiment without a succesful trial."""
    statuses = list(Trial.allowed_stati)
    statuses.remove("completed")

    exp = experiment_builder.build(name="test_single_exp")
    x = {"name": "/x", "type": "real"}

    x_value = 0
    for status in statuses:
        x["value"] = x_value
        trial = Trial(experiment=exp.id, params=[x], status=status)
        x_value += 1
        Database().write("trials", trial.to_dict())


@pytest.fixture
def single_with_trials(single_without_success):
    """Create an experiment with all types of trials."""
    exp = experiment_builder.build(name="test_single_exp")

    x = {"name": "/x", "type": "real", "value": 100}
    results = {"name": "obj", "type": "objective", "value": 0}
    trial = Trial(experiment=exp.id, params=[x], status="completed", results=[results])
    Database().write("trials", trial.to_dict())
    return exp.configuration


@pytest.fixture
def two_experiments(monkeypatch, storage):
    """Create an experiment and its child."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_double_exp",
            "./black_box.py",
            "--x~uniform(0,1)",
        ]
    )
    ensure_deterministic_id("test_double_exp", storage)

    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_double_exp",
            "--branch-to",
            "test_double_exp_child",
            "./black_box.py",
            "--x~+uniform(0,1,default_value=0)",
            "--y~+uniform(0,1,default_value=0)",
        ]
    )
    ensure_deterministic_id("test_double_exp_child", storage)


@pytest.fixture
def family_with_trials(two_experiments):
    """Create two related experiments with all types of trials."""
    exp = experiment_builder.build(name="test_double_exp")
    exp2 = experiment_builder.build(name="test_double_exp_child")
    x = {"name": "/x", "type": "real"}
    y = {"name": "/y", "type": "real"}

    x_value = 0
    for status in Trial.allowed_stati:
        x["value"] = x_value
        y["value"] = x_value
        trial = Trial(experiment=exp.id, params=[x], status=status)
        x["value"] = x_value
        trial2 = Trial(experiment=exp2.id, params=[x, y], status=status)
        x_value += 1
        Database().write("trials", trial.to_dict())
        Database().write("trials", trial2.to_dict())


@pytest.fixture
def unrelated_with_trials(family_with_trials, single_with_trials):
    """Create two unrelated experiments with all types of trials."""
    exp = experiment_builder.build(name="test_double_exp_child")

    Database().remove("trials", {"experiment": exp.id})
    Database().remove("experiments", {"_id": exp.id})


@pytest.fixture
def three_experiments(two_experiments, one_experiment):
    """Create a single experiment and an experiment and its child."""
    pass


@pytest.fixture
def three_experiments_with_trials(family_with_trials, single_with_trials):
    """Create three experiments, two unrelated, with all types of trials."""
    pass


@pytest.fixture
def three_experiments_family(two_experiments, storage):
    """Create three experiments, one of which is the parent of the other two."""
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_double_exp",
            "--branch-to",
            "test_double_exp_child2",
            "./black_box.py",
            "--x~+uniform(0,1,default_value=0)",
            "--z~+uniform(0,1,default_value=0)",
        ]
    )
    ensure_deterministic_id("test_double_exp_child2", storage)


@pytest.fixture
def three_family_with_trials(three_experiments_family, family_with_trials):
    """Create three experiments, all related, two direct children, with all types of trials."""
    exp = experiment_builder.build(name="test_double_exp_child2")
    x = {"name": "/x", "type": "real"}
    z = {"name": "/z", "type": "real"}

    x_value = 0
    for status in Trial.allowed_stati:
        x["value"] = x_value
        z["value"] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x, z], status=status)
        x_value += 1
        Database().write("trials", trial.to_dict())


@pytest.fixture
def three_experiments_family_branch(two_experiments, storage):
    """Create three experiments, each parent of the following one."""
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_double_exp_child",
            "--branch-to",
            "test_double_exp_grand_child",
            "./black_box.py",
            "--x~+uniform(0,1,default_value=0)",
            "--y~uniform(0,1,default_value=0)",
            "--z~+uniform(0,1,default_value=0)",
        ]
    )
    ensure_deterministic_id("test_double_exp_grand_child", storage)


@pytest.fixture
def three_family_branch_with_trials(
    three_experiments_family_branch, family_with_trials
):
    """Create three experiments, all related, one child and one grandchild,
    with all types of trials.

    """
    exp = experiment_builder.build(name="test_double_exp_grand_child")
    x = {"name": "/x", "type": "real"}
    y = {"name": "/y", "type": "real"}
    z = {"name": "/z", "type": "real"}

    x_value = 0
    for status in Trial.allowed_stati:
        x["value"] = x_value
        y["value"] = x_value * 10
        z["value"] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x, y, z], status=status)
        x_value += 1
        Database().write("trials", trial.to_dict())


@pytest.fixture
def two_experiments_same_name(one_experiment, storage):
    """Create two experiments with the same name but different versions."""
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_single_exp",
            "./black_box.py",
            "--x~uniform(0,1)",
            "--y~+normal(0,1)",
        ]
    )
    ensure_deterministic_id("test_single_exp", storage, version=2)


@pytest.fixture
def three_experiments_family_same_name(two_experiments_same_name, storage):
    """Create three experiments, two of them with the same name but different versions and one
    with a child.
    """
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_single_exp",
            "-v",
            "1",
            "-b",
            "test_single_exp_child",
            "./black_box.py",
            "--x~uniform(0,1)",
            "--y~+normal(0,1)",
        ]
    )
    ensure_deterministic_id("test_single_exp_child", storage)


@pytest.fixture
def three_experiments_branch_same_name(two_experiments_same_name, storage):
    """Create three experiments, two of them with the same name but different versions and last one
    with a child.
    """
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_single_exp",
            "-b",
            "test_single_exp_child",
            "./black_box.py",
            "--x~uniform(0,1)",
            "--y~normal(0,1)",
            "--z~+normal(0,1)",
        ]
    )
    ensure_deterministic_id("test_single_exp_child", storage)


@pytest.fixture
def three_experiments_same_name(two_experiments_same_name, storage):
    """Create three experiments with the same name but different versions."""
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_single_exp",
            "./black_box.py",
            "--x~uniform(0,1)",
            "--y~normal(0,1)",
            "--z~+normal(0,1)",
        ]
    )
    ensure_deterministic_id("test_single_exp", storage, version=3)


@pytest.fixture
def three_experiments_same_name_with_trials(two_experiments_same_name, storage):
    """Create three experiments with the same name but different versions."""
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "test_single_exp",
            "./black_box.py",
            "--x~uniform(0,1)",
            "--y~normal(0,1)",
            "--z~+normal(0,1)",
        ]
    )
    ensure_deterministic_id("test_single_exp", storage, version=3)

    exp = experiment_builder.build(name="test_single_exp", version=1)
    exp2 = experiment_builder.build(name="test_single_exp", version=2)
    exp3 = experiment_builder.build(name="test_single_exp", version=3)

    x = {"name": "/x", "type": "real"}
    y = {"name": "/y", "type": "real"}
    z = {"name": "/z", "type": "real"}
    x_value = 0
    for status in Trial.allowed_stati:
        x["value"] = x_value
        y["value"] = x_value * 10
        z["value"] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x], status=status)
        trial2 = Trial(experiment=exp2.id, params=[x, y], status=status)
        trial3 = Trial(experiment=exp3.id, params=[x, y, z], status=status)
        Database().write("trials", trial.to_dict())
        Database().write("trials", trial2.to_dict())
        Database().write("trials", trial3.to_dict())
        x_value += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.cmdline_parser`."""
import os
from collections import OrderedDict

import pytest

from orion.core.io.cmdline_parser import CmdlineParser


@pytest.fixture
def basic_config():
    """Return a simple configuration"""
    config = OrderedDict()

    config["_pos_0"] = "python"
    config["_pos_1"] = "script.py"
    config["_pos_2"] = "some"
    config["_pos_3"] = "pos"
    config["_pos_4"] = "args"
    config["with"] = "args"
    config["and"] = ["multiple", "args"]
    config["plus"] = True
    config["booleans"] = True
    config["equal"] = "value"

    return config


@pytest.fixture
def basic_keys():
    """Return keys of a simple configuration"""
    keys = OrderedDict()

    keys["_pos_0"] = "_pos_0"
    keys["_pos_1"] = "_pos_1"
    keys["_pos_2"] = "_pos_2"
    keys["_pos_3"] = "_pos_3"
    keys["_pos_4"] = "_pos_4"
    keys["with"] = "--with"
    keys["and"] = "--and"
    keys["plus"] = "--plus"
    keys["booleans"] = "--booleans"
    keys["equal"] = "--equal"

    return keys


@pytest.fixture
def to_format():
    """Return a commandline to format"""
    return "python 1 --arg value --args value1 value2 --boolean"


def test_key_to_arg():
    """Test the key to arg function"""
    cmdline_parser = CmdlineParser()
    assert cmdline_parser._key_to_arg("c") == "-c"
    assert cmdline_parser._key_to_arg("test") == "--test"
    assert cmdline_parser._key_to_arg("test-test") == "--test-test"


def test_parse_paths(monkeypatch):
    """Test the parse_paths function"""
    monkeypatch.chdir(os.path.dirname(__file__))
    cmdline_parser = CmdlineParser()
    assert cmdline_parser._parse_paths(__file__) == os.path.abspath(__file__)

    values = ["test_resolve_config.py", "test", "nada", __file__]
    parsed_values = cmdline_parser._parse_paths(values)
    assert parsed_values[0] == os.path.abspath("test_resolve_config.py")
    assert parsed_values[1] == "test"
    assert parsed_values[2] == "nada"
    assert parsed_values[3] == os.path.abspath(__file__)


def test_parse_arguments(basic_config, basic_keys):
    """Test the parsing of the commandline arguments"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "python script.py some pos args --with args --and multiple args "
        "--plus --booleans --equal=value".split(" ")
    )

    assert arguments == basic_config
    assert keys == basic_keys


def test_parse_arguments_template():
    """Test the creation of the template for arguments"""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(
        "python script.py some pos args "
        "--with args --and multiple args --plus --booleans --equal=value".split(" ")
    )

    assert cmdline_parser.template == [
        "{_pos_0}",
        "{_pos_1}",
        "{_pos_2}",
        "{_pos_3}",
        "{_pos_4}",
        "--with",
        "{with}",
        "--and",
        "{and[0]}",
        "{and[1]}",
        "--plus",
        "--booleans",
        "--equal",
        "{equal}",
    ]


def test_format(to_format):
    """Test that the format method assigns the correc values"""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(to_format.split(" "))

    keys, arguments = cmdline_parser._parse_arguments(to_format.split(" "))

    formatted = cmdline_parser.format(arguments)

    assert formatted == to_format.split(" ")


def test_parse_arguments_bad_command():
    """Test the fail cases of parsing"""
    cmdline_parser = CmdlineParser()

    with pytest.raises(ValueError) as exc_info:
        cmdline_parser.parse(
            "python script.py some pos args "
            "--with args --and multiple args --plus --booleans "
            "--and dummy.yaml".split(" ")
        )

    assert "Conflict: two arguments have the same name: and" in str(exc_info.value)


def test_has_already_been_parsed():
    """Test the template from the branching"""
    cmdline_parser = CmdlineParser()

    command = (
        "python script.py some pos args "
        "--with args --and multiple args --plus --booleans "
    )

    cmdline_parser.parse(command.split(" "))
    with pytest.raises(RuntimeError) as exc_info:
        cmdline_parser.parse(command.split(" "))

    assert "already" in str(exc_info.value)


def test_get_state_dict():
    """Test getting state dict."""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(
        "python script.py some pos args "
        "--with args --and multiple args --plus --booleans --equal=value".split(" ")
    )

    assert cmdline_parser.get_state_dict() == {
        "keys": list(map(list, cmdline_parser.keys.items())),
        "arguments": list(map(list, cmdline_parser.arguments.items())),
        "template": cmdline_parser.template,
    }


def test_set_state_dict():
    """Test that set_state_dict sets state properly to generate new cmdline."""
    cmdline_parser = CmdlineParser()

    cmdline_parser.set_state_dict(
        {
            "arguments": {},
            "template": [
                "{_pos_0}",
                "{_pos_1}",
                "--with",
                "{with}",
                "--plus",
                "--booleans",
            ],
        }
    )

    assert cmdline_parser.format(
        {"_pos_0": "voici", "_pos_1": "voila", "with": "classe"}
    ) == ["voici", "voila", "--with", "classe", "--plus", "--booleans"]


def test_parse_not_enough_dashes():
    """Test that arguments with many chars but one dash are supported even if it is not standard"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "pos -not-enough dashes " "--enough dashes -o my".split(" ")
    )

    assert arguments == OrderedDict(
        (("_pos_0", "pos"), ("not-enough", "dashes"), ("enough", "dashes"), ("o", "my"))
    )

    assert keys == OrderedDict(
        (
            ("_pos_0", "_pos_0"),
            ("not-enough", "-not-enough"),
            ("enough", "--enough"),
            ("o", "-o"),
        )
    )


def test_parse_fugly_underscores():
    """Test that underscores are kept as such no matter how fugly this is"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "pos -my_poor eyes --are_bleeding because --of-these underscores".split(" ")
    )

    assert arguments == OrderedDict(
        (
            ("_pos_0", "pos"),
            ("my_poor", "eyes"),
            ("are_bleeding", "because"),
            ("of-these", "underscores"),
        )
    )

    assert keys == OrderedDict(
        (
            ("_pos_0", "_pos_0"),
            ("my_poor", "-my_poor"),
            ("are_bleeding", "--are_bleeding"),
            ("of-these", "--of-these"),
        )
    )



"""
Benchmark Tasks definition
===========================
"""

from .base import BaseTask
from .branin import Branin
from .carromtable import CarromTable
from .eggholder import EggHolder
from .rosenbrock import RosenBrock

__all__ = ["BaseTask", "RosenBrock", "Branin", "CarromTable", "EggHolder"]



"""Perform functional tests merge of configuration levels"""
import copy
import datetime
import os
import random
from contextlib import contextmanager

import pytest
import yaml

import orion.core
import orion.core.cli
import orion.core.cli.hunt
import orion.core.evc.conflicts
import orion.core.io.resolve_config
from orion.client import get_experiment
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.singleton import SingletonNotInstantiatedError, update_singletons
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing.state import OrionState

script = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "..", "demo", "black_box.py"
)


class ConfigurationTestSuite:
    """Test suite for the configuration groups"""

    database = {}

    default_storage = {
        "type": "legacy",
        "database": {"type": "pickleddb", "host": "experiment.pkl"},
    }

    @contextmanager
    def setup_global_config(self, tmp_path):
        """Setup temporary yaml file for the global configuration"""
        with OrionState(storage=self.default_storage):
            conf_file = tmp_path / "config.yaml"
            conf_file.write_text(yaml.dump(self.config))
            conf_files = orion.core.DEF_CONFIG_FILES_PATHS
            orion.core.DEF_CONFIG_FILES_PATHS = [conf_file]
            orion.core.config = orion.core.build_config()
            try:
                yield conf_file
            finally:
                orion.core.DEF_CONFIG_FILES_PATHS = conf_files
                orion.core.config = orion.core.build_config()

    @contextmanager
    def setup_env_var_config(self, tmp_path):
        """Setup environment variables with temporary values"""
        with self.setup_global_config(tmp_path):
            tmp = {}
            for key, value in self.env_vars.items():
                tmp[key] = os.environ.pop(key, None)
                os.environ[key] = str(value)
            try:
                yield
            finally:
                for key, value in tmp.items():
                    if value:
                        os.environ[key] = str(value)
                    else:
                        del os.environ[key]

    @contextmanager
    def setup_db_config(self, tmp_path):
        """Setup database with temporary data"""
        with self.setup_env_var_config(tmp_path):
            storage = get_storage()
            storage.create_experiment(self.database)
            yield

    @contextmanager
    def setup_local_config(self, tmp_path):
        """Setup local configuration on top"""
        with self.setup_db_config(tmp_path):
            conf_file = tmp_path / "local.yaml"
            conf_file.write_text(yaml.dump(self.local))
            yield conf_file

    @contextmanager
    def setup_cmd_args_config(self, tmp_path):
        """Setup cmd args configuration... do nothing actually?"""
        with self.setup_local_config(tmp_path) as conf_file:
            yield conf_file

    def test_global_config(self, tmp_path, monkeypatch):
        """Test that global configuration is set properly based on global yaml"""
        update_singletons()
        self.sanity_check()
        with self.setup_global_config(tmp_path):
            self.check_global_config(tmp_path, monkeypatch)

    def test_env_var_config(self, tmp_path, monkeypatch):
        """Test that env vars are set properly in global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_env_var_config(tmp_path):
            self.check_env_var_config(tmp_path, monkeypatch)

    @pytest.mark.usefixtures(
        "with_user_userxyz", "version_XYZ", "mock_infer_versioning_metadata"
    )
    def test_db_config(self, tmp_path):
        """Test that exp config in db overrides global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_db_config(tmp_path):
            self.check_db_config()

    @pytest.mark.usefixtures("with_user_userxyz")
    def test_local_config(self, tmp_path, monkeypatch):
        """Test that local config overrides db/global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_local_config(tmp_path) as conf_file:
            self.check_local_config(tmp_path, conf_file, monkeypatch)

    @pytest.mark.usefixtures("with_user_userxyz", "version_XYZ")
    def test_cmd_args_config(self, tmp_path, monkeypatch):
        """Test that cmd_args config overrides local config"""
        update_singletons()
        self.sanity_check()
        with self.setup_cmd_args_config(tmp_path) as conf_file:
            self.check_cmd_args_config(tmp_path, conf_file, monkeypatch)


class TestStorage(ConfigurationTestSuite):
    """Test suite for storage configuration"""

    config = {
        "storage": {
            "type": "legacy",
            "database": {
                "name": "test_name",
                "type": "pickleddb",
                "host": "here.pkl",
                "port": 101,
            },
        }
    }

    env_vars = {
        "ORION_STORAGE_TYPE": "legacy",
        "ORION_DB_NAME": "test_env_var_name",
        "ORION_DB_TYPE": "pickleddb",
        "ORION_DB_ADDRESS": "there.pkl",
        "ORION_DB_PORT": "103",
    }

    local = {
        "storage": {
            "type": "legacy",
            "database": {"type": "pickleddb", "host": "local.pkl"},
        }
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.storage.to_dict() != self.config["storage"]

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == self.config["storage"]

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath("here.pkl")

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == {
            "type": self.env_vars["ORION_STORAGE_TYPE"],
            "database": {
                "name": self.env_vars["ORION_DB_NAME"],
                "type": self.env_vars["ORION_DB_TYPE"],
                "host": self.env_vars["ORION_DB_ADDRESS"],
                "port": int(self.env_vars["ORION_DB_PORT"]),
            },
        }

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath(self.env_vars["ORION_DB_ADDRESS"])

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == {
            "type": self.env_vars["ORION_STORAGE_TYPE"],
            "database": {
                "name": self.env_vars["ORION_DB_NAME"],
                "type": self.env_vars["ORION_DB_TYPE"],
                "host": self.env_vars["ORION_DB_ADDRESS"],
                "port": int(self.env_vars["ORION_DB_PORT"]),
            },
        }

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath("local.pkl")

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """No Storage config in cmdline, no test"""
        pass


class TestDatabaseDeprecated(ConfigurationTestSuite):
    """Test suite for deprecated database configuration."""

    config = {
        "database": {
            "name": "test_name",
            "type": "pickleddb",
            "host": "dbhere.pkl",
            "port": 101,
        }
    }

    env_vars = {
        "ORION_DB_NAME": "test_env_var_name",
        "ORION_DB_TYPE": "pickleddb",
        "ORION_DB_ADDRESS": "there.pkl",
        "ORION_DB_PORT": "103",
    }

    local = {"database": {"type": "pickleddb", "host": "dblocal.pkl"}}

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.database.to_dict() != self.config["database"]

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        update_singletons()

        assert orion.core.config.database.to_dict() == self.config["database"]

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath("dbhere.pkl")

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        update_singletons()

        assert orion.core.config.database.to_dict() == {
            "name": self.env_vars["ORION_DB_NAME"],
            "type": self.env_vars["ORION_DB_TYPE"],
            "host": self.env_vars["ORION_DB_ADDRESS"],
            "port": int(self.env_vars["ORION_DB_PORT"]),
        }

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath(self.env_vars["ORION_DB_ADDRESS"])

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        update_singletons()

        assert orion.core.config.database.to_dict() == {
            "name": self.env_vars["ORION_DB_NAME"],
            "type": self.env_vars["ORION_DB_TYPE"],
            "host": self.env_vars["ORION_DB_ADDRESS"],
            "port": int(self.env_vars["ORION_DB_PORT"]),
        }

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        command = f"hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == os.path.abspath("dblocal.pkl")

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """No Storage config in cmdline, no test"""
        pass


class TestExperimentConfig(ConfigurationTestSuite):
    """Test suite for experiment configuration"""

    config = {
        "experiment": {
            "max_trials": 10,
            "max_broken": 5,
            "working_dir": "here",
            "pool_size": 2,
            "worker_trials": 5,
            "algorithms": {"aa": {"b": "c", "d": {"e": "f"}}},
            "strategy": {"sa": {"c": "d", "e": {"f": "g"}}},
        }
    }

    env_vars = {
        "ORION_EXP_MAX_TRIALS": 20,
        "ORION_EXP_MAX_BROKEN": 12,
        "ORION_WORKING_DIR": "over_there",
    }

    database = {
        "name": "test-name",
        "version": 1,
        "max_trials": 75,
        "max_broken": 16,
        "working_dir": "in_db?",
        "algorithms": {"ab": {"d": "i", "f": "g"}},
        "producer": {"strategy": {"sb": {"e": "c", "d": "g"}}},
        "space": {"/x": "uniform(0, 1)"},
        "metadata": {
            "VCS": {
                "HEAD_sha": "test",
                "active_branch": None,
                "diff_sha": "diff",
                "is_dirty": False,
                "type": "git",
            },
            "datetime": datetime.datetime.utcnow(),
            "orion_version": "XYZ",
            "parser": {
                "cmd_priors": [["/x", "uniform(0, 1)"]],
                "config_file_data": {},
                "config_prefix": "config",
                "converter": None,
                "file_config_path": None,
                "file_priors": [],
                "parser": {
                    "arguments": [
                        ["_pos_0", "python"],
                        ["_pos_1", script],
                        ["x", "orion~uniform(0, 1)"],
                    ],
                    "keys": [["_pos_0", "_pos_0"], ["_pos_1", "_pos_1"], ["x", "-x"]],
                    "template": ["{_pos_0}", "{_pos_1}", "-x", "{x}"],
                },
            },
            "priors": {"/x": "uniform(0, 1)"},
            "user": "userxyz",
            "user_args": ["python", script, "-x~uniform(0, 1)"],
            "user_script": script,
        },
        "refers": {"adapter": [], "parent_id": None, "root_id": 1},
    }

    local = {
        "experiment": {
            "name": "test-name",
            "user": "useruvt",
            "max_trials": 50,
            "max_broken": 15,
            "working_dir": "here_again",
            "algorithms": {"ac": {"d": "e", "f": "g"}},
            "strategy": {"sd": {"b": "c", "d": "e"}},
        }
    }

    cmdargs = {
        "name": "exp-name",
        "user": "userabc",
        "version": 1,
        "exp-max-trials": 100,
        "exp-max-broken": 50,
        "working-dir": "cmdline-working-dir",
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()["experiment"] != self.config["experiment"]

    def _compare(self, base_config, experiment_config, ignore=tuple()):
        def _prune(config):
            config = copy.deepcopy(config)
            for key in ignore:
                config.pop(key, None)

            if "producer" in config:
                config["strategy"] = config.pop("producer")["strategy"]

            if "metadata" in config and "user" in config["metadata"]:
                config["user"] = config["metadata"]["user"]

            return config

        experiment_config = _prune(experiment_config)
        base_config = _prune(base_config)

        for key in list(experiment_config.keys()):
            if key not in base_config:
                experiment_config.pop(key)

        assert experiment_config == base_config

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        self._compare(
            self.config["experiment"], orion.core.config.to_dict()["experiment"]
        )

        command = f"hunt --init-only -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()

        experiment = get_experiment("test")
        self._compare(
            self.config["experiment"],
            experiment.configuration,
            ignore=["worker_trials"],
        )

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        assert (
            orion.core.config.experiment.max_trials
            == self.env_vars["ORION_EXP_MAX_TRIALS"]
        )
        assert (
            orion.core.config.experiment.max_broken
            == self.env_vars["ORION_EXP_MAX_BROKEN"]
        )
        assert (
            orion.core.config.experiment.working_dir
            == self.env_vars["ORION_WORKING_DIR"]
        )

        command = f"hunt --init-only -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        experiment = get_experiment("test")

        assert experiment.max_trials == self.env_vars["ORION_EXP_MAX_TRIALS"]
        assert experiment.max_broken == self.env_vars["ORION_EXP_MAX_BROKEN"]
        assert experiment.working_dir == self.env_vars["ORION_WORKING_DIR"]

    def check_db_config(self):
        """Check that db config overrides global/envvar config"""
        name = "test-name"
        command = f"hunt --worker-max-trials 0 -n {name}"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()

        experiment = get_experiment(name)
        self._compare(self.database, experiment.configuration, ignore=["worker_trials"])

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        command = f"hunt --worker-max-trials 0 -c {conf_file}"
        orion.core.cli.main(command.split(" "))

        storage = get_storage()

        experiment = get_experiment("test-name")
        self._compare(self.local["experiment"], experiment.configuration)

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        command = f"hunt --worker-max-trials 0 -c {conf_file} --branch-from test-name"
        command += " " + " ".join(
            "--{} {}".format(name, value) for name, value in self.cmdargs.items()
        )
        orion.core.cli.main(command.split(" "))

        storage = get_storage()

        experiment = get_experiment("exp-name")
        assert experiment.name == "exp-name"
        assert experiment.node.parent.name == "test-name"
        assert experiment.version == 1
        assert experiment.metadata["user"] == self.cmdargs["user"]
        assert experiment.max_trials == self.cmdargs["exp-max-trials"]
        assert experiment.max_broken == self.cmdargs["exp-max-broken"]
        assert experiment.working_dir == self.cmdargs["working-dir"]


class TestWorkerConfig(ConfigurationTestSuite):
    """Test suite for worker configuration"""

    config = {
        "worker": {
            "n_workers": 2,
            "executor": "dask",
            "executor_configuration": {"threads_per_worker": 1},
            "heartbeat": 30,
            "max_trials": 10,
            "max_broken": 5,
            "max_idle_time": 15,
            "interrupt_signal_code": 131,
            "user_script_config": "cfg",
        }
    }

    env_vars = {
        "ORION_N_WORKERS": 3,
        "ORION_EXECUTOR": "joblib",
        "ORION_HEARTBEAT": 40,
        "ORION_WORKER_MAX_TRIALS": 20,
        "ORION_WORKER_MAX_BROKEN": 6,
        "ORION_MAX_IDLE_TIME": 16,
        "ORION_INTERRUPT_CODE": 132,
        "ORION_USER_SCRIPT_CONFIG": "envcfg",
    }

    local = {
        "worker": {
            "n_workers": 4,
            "executor": "dask",
            "executor_configuration": {"threads_per_worker": 2},
            "heartbeat": 50,
            "max_trials": 30,
            "max_broken": 7,
            "max_idle_time": 17,
            "interrupt_signal_code": 133,
            "user_script_config": "lclcfg",
        }
    }

    cmdargs = {
        "n-workers": 1,
        "executor": "dask",
        "heartbeat": 70,
        "worker-max-trials": 1,
        "worker-max-broken": 8,
        "max-idle-time": 18,
        "interrupt-signal-code": 134,
        "user-script-config": "cmdcfg",
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()["worker"] != self.config["worker"]

    def _mock(self, monkeypatch):
        self._mock_exp_client(monkeypatch)
        self._mock_consumer(monkeypatch)
        self._mock_producer(monkeypatch)
        self._mock_workon(monkeypatch)

    def _mock_exp_client(self, monkeypatch):
        self.exp_client = None
        old_init = orion.client.experiment.ExperimentClient.__init__

        def init(c_self, *args, **kwargs):
            old_init(c_self, *args, **kwargs)
            self.exp_client = c_self

        monkeypatch.setattr(orion.client.experiment.ExperimentClient, "__init__", init)

    def _mock_consumer(self, monkeypatch):
        self.consumer = None
        old_init = orion.core.cli.hunt.Consumer.__init__

        def init(c_self, *args, **kwargs):
            old_init(c_self, *args, **kwargs)
            self.consumer = c_self

        monkeypatch.setattr(orion.core.cli.hunt.Consumer, "__init__", init)

    def _mock_producer(self, monkeypatch):
        self.producer = None
        old_init = orion.core.cli.hunt.Producer.__init__

        def init(p_self, *args, **kwargs):
            old_init(p_self, *args, **kwargs)
            self.producer = p_self

        monkeypatch.setattr(orion.core.cli.hunt.Producer, "__init__", init)

    def _mock_workon(self, monkeypatch):
        workon = orion.core.cli.hunt.workon

        self.workon_kwargs = None

        def mocked_workon(experiment, **kwargs):
            self.workon_kwargs = kwargs
            return workon(experiment, **kwargs)

        monkeypatch.setattr("orion.core.cli.hunt.workon", mocked_workon)

    def _check_mocks(self, config):
        self._check_exp_client(config)
        self._check_consumer(config)
        self._check_producer(config)
        self._check_workon(config)

    def _check_exp_client(self, config):
        assert self.exp_client.heartbeat == config["heartbeat"]

    def _check_consumer(self, config):
        assert (
            self.consumer.template_builder.config_prefix == config["user_script_config"]
        )
        assert self.consumer.interrupt_signal_code == config["interrupt_signal_code"]

    def _check_producer(self, config):
        assert self.producer.max_idle_time == config["max_idle_time"]

    def _check_workon(self, config):
        assert self.workon_kwargs["n_workers"] == config["n_workers"]
        assert self.workon_kwargs["executor"] == config["executor"]
        assert (
            self.workon_kwargs["executor_configuration"]
            == config["executor_configuration"]
        )
        assert self.workon_kwargs["max_trials"] == config["max_trials"]
        assert self.workon_kwargs["max_broken"] == config["max_broken"]

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        assert orion.core.config.to_dict()["worker"] == self.config["worker"]

        self._mock(monkeypatch)

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        self._check_mocks(self.config["worker"])

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        env_var_config = {
            "n_workers": self.env_vars["ORION_N_WORKERS"],
            "executor": self.env_vars["ORION_EXECUTOR"],
            "executor_configuration": self.config["worker"]["executor_configuration"],
            "heartbeat": self.env_vars["ORION_HEARTBEAT"],
            "max_trials": self.env_vars["ORION_WORKER_MAX_TRIALS"],
            "max_broken": self.env_vars["ORION_WORKER_MAX_BROKEN"],
            "max_idle_time": self.env_vars["ORION_MAX_IDLE_TIME"],
            "interrupt_signal_code": self.env_vars["ORION_INTERRUPT_CODE"],
            "user_script_config": self.env_vars["ORION_USER_SCRIPT_CONFIG"],
        }

        assert orion.core.config.to_dict()["worker"] == env_var_config

        # Override executor configuration otherwise joblib will fail.
        orion.core.config.worker.executor_configuration = {}
        env_var_config["executor_configuration"] = {}

        self._mock(monkeypatch)

        command = f"hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        self._check_mocks(env_var_config)

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        self._mock(monkeypatch)

        # Override executor so that executor and configuration are coherent in global config
        os.environ["ORION_EXECUTOR"] = "dask"

        command = f"hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        self._check_mocks(self.local["worker"])

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        config = {
            "n_workers": self.cmdargs["n-workers"],
            "executor": self.cmdargs["executor"],
            "executor_configuration": {"threads_per_worker": 2},
            "heartbeat": self.cmdargs["heartbeat"],
            "max_trials": self.cmdargs["worker-max-trials"],
            "max_broken": self.cmdargs["worker-max-broken"],
            "max_idle_time": self.cmdargs["max-idle-time"],
            "interrupt_signal_code": self.cmdargs["interrupt-signal-code"],
            "user_script_config": self.cmdargs["user-script-config"],
        }

        self._mock(monkeypatch)

        # Override executor so that executor and configuration are coherent in global config
        os.environ["ORION_EXECUTOR"] = "dask"

        command = f"hunt --worker-max-trials 0 -c {conf_file} -n cmd-test"
        command += " " + " ".join(
            "--{} {}".format(name, value) for name, value in self.cmdargs.items()
        )
        command += f" python {script} -x~uniform(0,1)"
        orion.core.cli.main(command.split(" "))

        self._check_mocks(config)


class TestEVCConfig(ConfigurationTestSuite):
    """Test for EVC configuration"""

    config = {
        "evc": {
            "auto_resolution": False,
            "manual_resolution": True,
            "non_monitored_arguments": ["test", "one"],
            "ignore_code_changes": True,
            "algorithm_change": True,
            "code_change_type": "noeffect",
            "cli_change_type": "noeffect",
            "config_change_type": "noeffect",
            "orion_version_change": True,
        }
    }

    env_vars = {
        "ORION_EVC_MANUAL_RESOLUTION": "",
        "ORION_EVC_NON_MONITORED_ARGUMENTS": "test:two:others",
        "ORION_EVC_IGNORE_CODE_CHANGES": "",
        "ORION_EVC_ALGO_CHANGE": "",
        "ORION_EVC_CODE_CHANGE": "unsure",
        "ORION_EVC_CMDLINE_CHANGE": "unsure",
        "ORION_EVC_CONFIG_CHANGE": "unsure",
        "ORION_EVC_ORION_VERSION_CHANGE": "",
    }

    local = {
        "evc": {
            "manual_resolution": True,
            "non_monitored_arguments": ["test", "local"],
            "ignore_code_changes": True,
            "algorithm_change": True,
            "code_change_type": "break",
            "cli_change_type": "break",
            "config_change_type": "noeffect",
            "orion_version_change": True,
        }
    }

    cmdargs = {
        "manual-resolution": False,
        "non-monitored-arguments": "test:cmdargs",
        "ignore-code-changes": False,
        "algorithm-change": False,
        "code-change-type": "noeffect",
        "cli-change-type": "unsure",
        "config-change-type": "break",
        "orion-version-change": False,
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()["evc"] != self.config["evc"]

    def _mock_consumer(self, monkeypatch):
        self.consumer = None
        old_init = orion.core.cli.hunt.Consumer.__init__

        def init(c_self, *args, **kwargs):
            old_init(c_self, *args, **kwargs)
            self.consumer = c_self

        monkeypatch.setattr(orion.core.cli.hunt.Consumer, "__init__", init)

    def _check_consumer(self, config):
        assert self.consumer.ignore_code_changes == config["ignore_code_changes"]

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        assert orion.core.config.to_dict()["evc"] == self.config["evc"]

        name = "global-test"
        command = (
            f"hunt --worker-max-trials 0 -n {name} python {script} -x~uniform(0,1)"
        )
        assert orion.core.cli.main(command.split(" ")) == 0

        # Test that manual_resolution is True and it branches when changing cli
        assert orion.core.cli.main(f"{command} --cli-change ".split(" ")) == 1

        command = "hunt --auto-resolution " + command[5:]

        command = self._check_cli_change(
            name, command, version=1, change_type="noeffect"
        )
        self._check_non_monitored_arguments(
            name, command, version=2, non_monitored_arguments=["test", "one"]
        )
        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type="noeffect"
        )
        self._check_code_change(
            monkeypatch,
            name,
            command,
            version=3,
            mock_ignore_code_changes=None,
            ignore_code_changes=self.config["evc"]["ignore_code_changes"],
            change_type=self.config["evc"]["code_change_type"],
        )

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        assert not orion.core.config.evc.manual_resolution
        assert not orion.core.config.evc.ignore_code_changes
        assert not orion.core.config.evc.algorithm_change
        assert orion.core.config.evc.non_monitored_arguments == [
            "test",
            "two",
            "others",
        ]
        assert orion.core.config.evc.code_change_type == "unsure"
        assert orion.core.config.evc.cli_change_type == "unsure"
        assert orion.core.config.evc.config_change_type == "unsure"
        assert not orion.core.config.evc.orion_version_change

        name = "env-var-test"
        command = (
            f"hunt --worker-max-trials 0 -n {name} python {script} -x~uniform(0,1)"
        )
        assert orion.core.cli.main(command.split(" ")) == 0

        # TODO: Anything to test still???
        command = self._check_cli_change(name, command, version=1, change_type="unsure")
        command = self._check_non_monitored_arguments(
            name, command, version=2, non_monitored_arguments=["test", "two", "others"]
        )
        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type="unsure"
        )

        self._check_code_change(
            monkeypatch,
            name,
            command,
            version=3,
            mock_ignore_code_changes=None,
            ignore_code_changes=bool(self.env_vars["ORION_EVC_IGNORE_CODE_CHANGES"]),
            change_type=self.env_vars["ORION_EVC_CODE_CHANGE"],
        )

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        name = "local-test"
        command = (
            f"hunt --worker-max-trials 0 -n {name} -c {conf_file} "
            f"python {script} -x~uniform(0,1)"
        )

        assert orion.core.cli.main(command.split(" ")) == 0

        # Test that manual_resolution is True and it branches when changing cli
        assert orion.core.cli.main(f"{command} --cli-change ".split(" ")) == 1

        command = "hunt --auto-resolution " + command[5:]

        command = self._check_cli_change(
            name, command, version=1, change_type=self.local["evc"]["cli_change_type"]
        )
        command = self._check_non_monitored_arguments(
            name,
            command,
            version=2,
            non_monitored_arguments=self.local["evc"]["non_monitored_arguments"],
        )
        self._check_script_config_change(
            tmp_path,
            name,
            command,
            version=2,
            change_type=self.local["evc"]["config_change_type"],
        )
        self._check_code_change(
            monkeypatch,
            name,
            command,
            version=3,
            mock_ignore_code_changes=True,
            ignore_code_changes=self.local["evc"]["ignore_code_changes"],
            change_type=self.local["evc"]["code_change_type"],
        )

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        name = "cmd-test"
        command = (
            f"hunt --worker-max-trials 0 -c {conf_file} -n {name} "
            "--auto-resolution "
            "--non-monitored-arguments test:cmdargs "
            "--code-change-type noeffect "
            "--cli-change-type unsure "
            "--config-change-type break "
            f"python {script} -x~uniform(0,1)"
        )
        assert orion.core.cli.main(command.split(" ")) == 0

        command = self._check_cli_change(
            name, command, version=1, change_type=self.cmdargs["cli-change-type"]
        )
        command = self._check_non_monitored_arguments(
            name,
            command,
            version=2,
            non_monitored_arguments=self.cmdargs["non-monitored-arguments"].split(":"),
        )

        self._check_script_config_change(
            tmp_path,
            name,
            command,
            version=2,
            change_type=self.cmdargs["config-change-type"],
        )

        # Mock local to ignore_code_changes=False
        fetch_config = orion.core.io.resolve_config.fetch_config

        def mock_local(cmdargs):
            config = fetch_config(cmdargs)
            config["evc"]["ignore_code_changes"] = False
            return config

        monkeypatch.setattr(orion.core.io.resolve_config, "fetch_config", mock_local)

        # Check that ignore_code_changes is rightly False
        self._check_code_change(
            monkeypatch,
            name,
            command,
            version=3,
            mock_ignore_code_changes=False,
            ignore_code_changes=False,
            change_type=self.cmdargs["code-change-type"],
        )

        command = "hunt --ignore-code-changes " + command[5:]

        # Check that ignore_code_changes is now True
        self._check_code_change(
            monkeypatch,
            name,
            command,
            version=4,
            mock_ignore_code_changes=True,
            ignore_code_changes=True,
            change_type=self.cmdargs["code-change-type"],
        )

    def _check_cli_change(self, name, command, version, change_type):
        command += " --cli-change"

        # Test that manual_resolution is False and it branches when changing cli
        assert orion.core.cli.main(command.split(" ")) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        assert experiment.refers["adapter"].configuration[0] == {
            "of_type": "commandlinechange",
            "change_type": change_type,
        }

        return command

    def _check_non_monitored_arguments(
        self, name, command, version, non_monitored_arguments
    ):
        for argument in non_monitored_arguments:
            command += f" --{argument} "

        # Test that cli change with non-monitored args do not cause branching
        assert orion.core.cli.main(command.split(" ")) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version

        return command

    def _check_code_change(
        self,
        monkeypatch,
        name,
        command,
        version,
        mock_ignore_code_changes,
        ignore_code_changes,
        change_type,
    ):

        # Test that code change is handled with 'no-effect'
        def fixed_dictionary(user_script):
            """Create VCS"""
            vcs = {}
            vcs["type"] = "git"
            vcs["is_dirty"] = False
            vcs["HEAD_sha"] = "test " + str(random.random())
            vcs["active_branch"] = None
            vcs["diff_sha"] = "diff"
            return vcs

        monkeypatch.setattr(
            orion.core.io.resolve_config, "infer_versioning_metadata", fixed_dictionary
        )

        self._mock_consumer(monkeypatch)

        detect = orion.core.evc.conflicts.CodeConflict.detect

        def mock_detect(old_config, new_config, branching_config=None):
            if branching_config and "ignore_code_changes" in branching_config:
                assert (
                    branching_config["ignore_code_changes"] is mock_ignore_code_changes
                )
                branching_config["ignore_code_changes"] = False
            return detect(old_config, new_config, branching_config)

        monkeypatch.setattr(
            orion.core.evc.conflicts.CodeConflict, "detect", mock_detect
        )
        assert orion.core.cli.main(command.split(" ")) == 0
        self._check_consumer({"ignore_code_changes": ignore_code_changes})

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        assert experiment.refers["adapter"].configuration[0] == {
            "of_type": "codechange",
            "change_type": change_type,
        }

        monkeypatch.undo()

    def _check_script_config_change(
        self, tmp_path, name, command, version, change_type
    ):

        # Test that config change is handled with 'break'
        with self.setup_user_script_config(tmp_path) as user_script_config:

            command += f" --config {user_script_config}"
            assert orion.core.cli.main(command.split(" ")) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        print(experiment.refers["adapter"].configuration)
        assert len(experiment.refers["adapter"].configuration) == 2
        assert experiment.refers["adapter"].configuration[1] == {
            "of_type": "scriptconfigchange",
            "change_type": change_type,
        }

    @contextmanager
    def setup_user_script_config(self, tmp_path):
        """Setup temporary dummy user script config"""
        conf_file = tmp_path / "user_script_config.yaml"
        config = {"what": "ever"}
        conf_file.write_text(yaml.dump(config))
        yield conf_file



# -*- coding: utf-8 -*-
"""
Base class and function utilities for cli
=========================================
"""
import argparse
import logging
import sys
import textwrap

import orion
from orion.core.io.database import DatabaseError
from orion.core.utils.exceptions import (
    BranchingEvent,
    InexecutableUserScript,
    MissingResultFile,
    NoConfigurationError,
    NoNameError,
)

CLI_DOC_HEADER = "Oríon CLI for asynchronous distributed optimization"


class OrionArgsParser:
    """Parser object handling the upper-level parsing of Oríon's arguments."""

    def __init__(self, description=CLI_DOC_HEADER):
        """Create the pre-command arguments"""
        self.description = description

        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent(description),
        )

        self.parser.add_argument(
            "-V",
            "--version",
            action="version",
            version="orion " + orion.core.__version__,
        )

        self.parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="logging levels of information about the process (-v: INFO. -vv: DEBUG)",
        )

        self.parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="Use debugging mode with EphemeralDB.",
        )

        self.subparsers = self.parser.add_subparsers(dest="command")

    def get_subparsers(self):
        """Return the subparser object for this parser."""
        return self.subparsers

    def parse(self, argv):
        """Call argparse and generate a dictionary of arguments' value"""
        args = vars(self.parser.parse_args(argv))

        verbose = args.pop("verbose", 0)
        levels = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}
        logging.basicConfig(
            format="%(asctime)-15s::%(levelname)s::%(name)s::%(message)s",
            level=levels.get(verbose, logging.DEBUG),
        )

        if args["command"] is None:
            self.parser.parse_args(["--help"])

        function = args.pop("func", None)
        empty_command = (argv[-1] if argv else sys.argv[-1]) == args["command"]
        if function is None or (empty_command and args.pop("help_empty", False)):
            self.parser.parse_args([args["command"], "--help"])

        return args, function

    def execute(self, argv):
        """Execute main function of the subparser"""
        try:
            args, function = self.parse(argv)
            returncode = function(args)
        except (
            NoConfigurationError,
            NoNameError,
            DatabaseError,
            MissingResultFile,
            BranchingEvent,
            InexecutableUserScript,
        ) as e:
            print("Error:", e, file=sys.stderr)

            if args.get("verbose", 0) >= 2:
                raise e

            return 1

        except KeyboardInterrupt:
            print("Orion is interrupted.")
            return 130

        return 0 if returncode is None else returncode


def get_basic_args_group(
    parser,
    group_name="Oríon arguments",
    group_help="These arguments determine orion's behaviour",
):
    """Return the basic arguments for any command."""
    basic_args_group = parser.add_argument_group(group_name, description=group_help)

    basic_args_group.add_argument(
        "-n",
        "--name",
        type=str,
        metavar="stringID",
        help="experiment's unique name; "
        "(default: None - specified either here or in a config)",
    )

    basic_args_group.add_argument(
        "-u",
        "--user",
        type=str,
        help="user associated to experiment's unique name; "
        "(default: $USER - can be overriden either here or in a config)",
    )

    basic_args_group.add_argument(
        "-v",
        "--version",
        type=int,
        help="specific version of experiment to fetch; "
        "(default: None - latest experiment.)",
    )

    basic_args_group.add_argument(
        "-c",
        "--config",
        type=argparse.FileType("r"),
        metavar="path-to-config",
        help="user provided " "orion configuration file",
    )

    return basic_args_group


def get_user_args_group(parser):
    """
    Return the user group arguments for any command.
    User group arguments are composed of the user script and the user args
    """
    usergroup = parser.add_argument_group(
        "User script related arguments",
        description="These arguments determine user's script behaviour "
        "and they can serve as orion's parameter declaration.",
    )

    usergroup.add_argument(
        "user_args",
        nargs=argparse.REMAINDER,
        metavar="...",
        help="Command line of user script. A configuration "
        "file intended to be used with 'userscript' must be given as a path "
        "in the **first positional** argument OR using `--config=<path>` "
        "keyword argument.",
    )

    return usergroup



# -*- coding: utf-8 -*-
"""
=================
Python API basics
=================

This short tutorial will show you the basics to use Oríon in python. We will optimize a simple
1-d ``rosenbrock`` function with random search and TPE and visualize the regret curve to compare
the algorithms.

We first import the only function needed, :func:`build experiment <orion.client.build_experiment>`.
"""
from orion.client import build_experiment

#%%
# We configure the database with PickledDB so that the results are saved locally on disk. This
# enables resuming the experiment and running parallel workers.
#
# .. TODO Replace with just nothing when pickleddb becomes the default.
#

storage = {
    "type": "legacy",
    "database": {
        "type": "pickleddb",
        "host": "./db.pkl",
    },
}

#%%
# We define the search space for the optimization. Here, the optimization algorithm may explore
# real values for ``x`` between 0 and 30 only. See documentation of :ref:`search-space` for more
# information.

space = {"x": "uniform(0, 30)"}

#%%
# We then build the experiment with the name ``random-rosenbrock``. The name is by Oríon as
# an `id` for the experiment. Each experiment must have a unique name.

experiment = build_experiment(
    "random-rosenbrock",
    space=space,
    storage=storage,
)

#%%
# For this example we use a 1-d rosenbrock function. We must return a list of results,
# for Oríon. Results must have the format
# ``{name: <str>: type: <'objective', 'constraint' or 'gradient'>, value=<float>}`` otherwise
# a ``ValueError`` will be raised. At least one of the results must have the type ``objective``,
# the metric that is minimized by the algorithm.


def rosenbrock(x, noise=None):
    """Evaluate partial information of a quadratic."""
    y = x - 34.56789
    z = 4 * y ** 2 + 23.4

    return [{"name": "objective", "type": "objective", "value": z}]


#%%
# We then pass the function ``rosenbrock`` to
# :meth:`workon() <orion.client.experiment.ExperimentClient.workon>`. This method
# will iteratively try new sets of hyperparameters suggested by the optimization algorithm
# until it reaches 20 trials.

experiment.workon(rosenbrock, max_trials=20)

#%%
# Now let's plot the regret curve to see how well went the optimization.

experiment.plot.regret().show()

#%%
# .. This file is produced by docs/scripts/build_database_and_plots.py
#
# .. raw:: html
#     :file: ../_static/random-rosenbrock_regret.html
#
# We have here on the x-axis the trials order by suggestion time, that is the order they
# were tried inside the ``workon`` loop. The y-axis is the objective of the trials, the lower the
# better. The red curve represents the regret, the best achieved objective at time `x`, and the
# blue dots are the objectives of the different trials.
#
# Fortunately, we can see looking at the regret curve that the best results converge close to the
# minimum. That means the optimization was at least reasonably good. On the other-hand, looking at
# the trials we see that the algorithm made suggestions far from optimal from the beginning
# to the end of the optimization. That is to be expected, by default Oríon uses random search. All
# trials were randomly sampled. Let's try next a different algorithm to observe a proper convergence
# behavior.
#
# Let's use a
# `Tree Parzen Estimator
# <https://papers.nips.cc/paper/2011/hash/86e8f7ab32cfd12577bc2619bc635690-Abstract.html>`_
# that can easily find the optimal solution. We specify the algorithm configuration t
# :func:`build experiment <orion.client.build_experiment>`

experiment = build_experiment(
    "tpe-rosenbrock",
    space=space,
    algorithms={"tpe": {"n_initial_points": 5}},
    storage=storage,
)

#%%
# We then again run the optimization for 20 trials and plot the regret.

experiment.workon(rosenbrock, max_trials=20)

experiment.plot.regret().show()

# sphinx_gallery_thumbnail_path = '_static/python.png'

#%%
# .. This file is produced by docs/scripts/build_database_and_plots.py
#
# .. raw:: html
#     :file: ../_static/tpe-rosenbrock_regret.html
#
# We can see the stark difference with the results of the random search. After the first 5 random
# points (see tpe's configuration above), the TPE already identified the optimal region for `x`
# and explored this subspace.
#
# Further readings
# ----------------
#
# User guides:
#
# - :ref:`Search space <search-space>`
# - :ref:`Algorithms <Setup Algorithms>`
# - :ref:`Visualizations <visualizations>`
#
# API:
#
# - :func:`orion.client.build_experiment`
# - :func:`orion.client.get_experiment`
# - :class:`orion.client.experiment.ExperimentClient`



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for db set."""
import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage


def execute(command, assert_code=0):
    """Execute orion command and return returncode"""
    returncode = orion.core.cli.main(command.split(" "))
    assert returncode == assert_code


def test_no_exp(storage, capsys):
    """Test that set non-existing exp exits gracefully"""
    execute("db set i-dont-exist whatever=1 idontcare=2", assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith(
        "Error: No experiment with given name 'i-dont-exist'"
    )


def test_confirm_name(monkeypatch, single_with_trials):
    """Test name must be confirmed for update"""

    def incorrect_name(*args):
        return "oops"

    monkeypatch.setattr("builtins.input", incorrect_name)

    execute("db set test_single_exp status=broken status=interrupted", assert_code=1)

    def correct_name(*args):
        return "test_single_exp"

    monkeypatch.setattr("builtins.input", correct_name)

    assert len(get_storage()._fetch_trials({"status": "broken"})) > 0
    execute("db set test_single_exp status=broken status=interrupted")
    assert len(get_storage()._fetch_trials({"status": "broken"})) == 0


def test_invalid_query(single_with_trials, capsys):
    """Test error message when query is invalid"""
    execute("db set -f test_single_exp whatever=1 idontcare=2", assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: Invalid query attribute `whatever`.")


def test_invalid_update(single_with_trials, capsys):
    """Test error message when update attribute is invalid"""
    execute("db set -f test_single_exp status=new yoopidoo=2", assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: Invalid update attribute `yoopidoo`.")


def test_update_trial(single_with_trials, capsys):
    """Test that trial is updated properly"""
    trials = get_storage()._fetch_trials({})
    assert sum(trial.status == "broken" for trial in trials) > 0
    trials = dict(zip((trial.id for trial in trials), trials))
    execute("db set -f test_single_exp status=broken status=interrupted")
    for trial in get_storage()._fetch_trials({}):
        if trials[trial.id].status == "broken":
            assert trial.status == "interrupted", "status not changed properly"
        else:
            assert (
                trials[trial.id].status == trial.status
            ), "status should not have been changed"

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_update_trial_with_id(single_with_trials, capsys):
    """Test that trial is updated properly when querying with the id"""
    trials = get_storage()._fetch_trials({})
    trials = dict(zip((trial.id for trial in trials), trials))
    trial = get_storage()._fetch_trials({"status": "broken"})[0]
    execute(f"db set -f test_single_exp id={trial.id} status=interrupted")
    for new_trial in get_storage()._fetch_trials({}):
        if new_trial.id == trial.id:
            assert new_trial.status == "interrupted", "status not changed properly"
        else:
            status_unchanged = trials[new_trial.id].status == new_trial.status
            assert status_unchanged, "status should not have been changed"

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_update_no_match_query(single_with_trials, capsys):
    """Test that no trials are updated when there is no match"""
    trials = get_storage()._fetch_trials({})
    trials = dict(zip((trial.id for trial in trials), trials))
    execute("db set -f test_single_exp status=invalid status=interrupted")
    for trial in get_storage()._fetch_trials({}):
        assert (
            trials[trial.id].status == trial.status
        ), "status should not have been changed"

    captured = capsys.readouterr()

    assert captured.out.endswith("0 trials modified\n")


def test_update_child_only(three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.load(name="test_single_exp", version=1)
    exp2 = experiment_builder.load(name="test_single_exp", version=2)
    exp3 = experiment_builder.load(name="test_single_exp", version=3)
    execute("db set -f test_single_exp --version 2 status=broken status=interrupted")
    assert len(exp1.fetch_trials_by_status("broken")) > 0
    assert len(exp2.fetch_trials_by_status("broken")) == 0
    assert len(exp3.fetch_trials_by_status("broken")) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("2 trials modified\n")


def test_update_default_leaf(three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.load(name="test_single_exp", version=1)
    exp2 = experiment_builder.load(name="test_single_exp", version=2)
    exp3 = experiment_builder.load(name="test_single_exp", version=3)
    execute("db set -f test_single_exp status=broken status=interrupted")
    assert len(exp1.fetch_trials_by_status("broken")) > 0
    assert len(exp2.fetch_trials_by_status("broken")) > 0
    assert len(exp3.fetch_trials_by_status("broken")) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_no_update_id_from_parent(three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.load(name="test_single_exp", version=1)
    exp2 = experiment_builder.load(name="test_single_exp", version=2)
    exp3 = experiment_builder.load(name="test_single_exp", version=3)
    trial = exp1.fetch_trials_by_status("broken")[0]
    execute(f"db set -f test_single_exp id={trial.id} status=shouldnothappen")
    assert len(exp1.fetch_trials_by_status("shouldnothappen")) == 0
    assert len(exp2.fetch_trials_by_status("shouldnothappen")) == 0
    assert len(exp3.fetch_trials_by_status("shouldnothappen")) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("0 trials modified\n")



# -*- coding: utf-8 -*-
"""Generic tests for Algorithms"""
import copy
import functools
import inspect
import itertools
from collections import defaultdict

import numpy
import pytest

import orion.algo.base
from orion.algo.asha import ASHA
from orion.algo.gridsearch import GridSearch
from orion.algo.hyperband import Hyperband
from orion.algo.random import Random
from orion.algo.tpe import TPE
from orion.benchmark.task.branin import Branin
from orion.core.io.space_builder import SpaceBuilder
from orion.core.worker.primary_algo import PrimaryAlgo
from orion.testing.space import build_space

algorithms = {
    "evolutionaryes": "what is the config?",
}


def recursive_getattr(obj, attribute):
    """Get attribute recursively based on str with pattern sub.(sub.)*name"""
    attributes = attribute.split(".")

    attribute = getattr(obj, attributes[0])
    if len(attributes) == 1:
        return attribute

    return recursive_getattr(attribute, ".".join(attributes[1:]))


def spy_attr(mocker, algo, attribute):
    """Return a mocker.spy object on the algorithms's given `attribute`"""
    attributes = attribute.split(".")
    attr_to_mock = attributes[-1]
    if len(attributes) > 1:
        obj = recursive_getattr(algo, ".".join(attributes[:-1]))
    else:
        obj = algo
    return mocker.spy(obj, attr_to_mock)


methods_with_phase = defaultdict(set)


phase_docstring = """\
This test is parametrizable with phases.
See ``orion.testing.algo.BaseAlgoTests.set_phases``.\
"""


def phase(method):
    """Decorator to mark methods that must be parametrized with phases."""
    class_name = ".".join(method.__qualname__.split(".")[:-1])
    methods_with_phase[class_name].add(method.__name__)

    if method.__doc__ is None:
        method.__doc__ = phase_docstring
    else:
        method.__doc__ += "\n\n" + phase_docstring

    return method


def parametrize_this(cls, method_name, attrs, ids):
    """Parametrize a method with phases.

    Notes
    -----
    We need to replace the method to avoid parametrizing the base one.
    This is a fugly hack because of pytest limitations with inheritance for test classes.
    """

    base_method = getattr(cls, method_name)

    def method(self, mocker, num, attr):
        print(method_name, base_method)
        return base_method(self, mocker, num, attr)

    setattr(cls, method_name, method)

    pytest.mark.parametrize("num,attr", attrs, ids=ids)(method)


def customized_mutate_example(search_space, rng, old_value, **kwargs):
    """Define a customized mutate function example"""
    multiply_factor = kwargs.pop("multiply_factor", 3.0)
    add_factor = kwargs.pop("add_factor", 1)
    if search_space.type == "real":
        new_value = old_value / multiply_factor
    elif search_space.type == "integer":
        new_value = int(old_value + add_factor)
    else:
        new_value = old_value
    return new_value


class BaseAlgoTests:
    """Generic Test-suite for HPO algorithms.

    This test-suite covers all typical cases for HPO algorithms. To use it for a new algorithm,
    the class inheriting from this one must redefine the attributes ``algo_name`` with
    the name of the algorithm used to create it with the algorithm factory
    ``orion.core.worker.primary_algo.PrimaryAlgo`` and ``config`` with a base configuration for the
    algorithm that contains all its arguments. The base space can be redefine if needed
    with the attribute ``space``.

    Many algorithms have different phases that should be tested. For instance
    TPE have a first phase of random search and a second of Bayesian Optimization.
    The random search and Bayesian optimization are different implementations and both should be
    tested. For this reason, the class method ``orion.testing.algo.BaseAlgoTests.set_phases`` must
    be called to parametrize the tests with phases. Failure to doing so will causes the tests to
    crash. See ``tests/unittests/algo/test_tpe.py`` for an example.
    """

    algo_name = None
    config = {}
    max_trials = 200
    space = {"x": "uniform(0, 1)", "y": "uniform(0, 1)"}

    @classmethod
    def set_phases(cls, phases):
        """Parametrize the tests with different phases.

        Some algorithms have different phases that should be tested. For instance
        TPE have a first phase of random search and a second of Bayesian Optimization.
        The random search and Bayesian optimization are different implementations and both should be
        tested.

        Parameters
        ----------
        phases: list of tuples
            The different phases to test. The format of the tuples should be
            (str(id of the test),
            int(number of trials before the phase begins),
            str(name of the algorithm's attribute to spy (ex: "space.sample"))
            )
        """
        ids = [phase[0] for phase in phases]
        attrs = [phase[1:] for phase in phases]

        cls_methods_with_phase = (
            methods_with_phase["BaseAlgoTests"] | methods_with_phase[cls.__name__]
        )

        for method_name in sorted(cls_methods_with_phase):
            parametrize_this(cls, method_name, attrs, ids)

    def create_algo(self, config=None, space=None, **kwargs):
        """Create the algorithm based on config.

        Parameters
        ----------
        config: dict, optional
            The configuration for the algorithm. ``self.config`` will be used
            if ``config`` is ``None``.
        space: ``orion.algo.space.Space``, optional
            Space object to pass to algo. The output of ``self.create_space()``
            will be used if ``space`` is ``None``.
        kwargs: dict
            Values to override algorithm configuration.
        """
        config = copy.deepcopy(config or self.config)
        config.update(kwargs)
        algo = PrimaryAlgo(
            space or self.create_space(),
            {self.algo_name: config},
        )
        algo.algorithm.max_trials = self.max_trials
        return algo

    def update_space(self, test_space):
        """Get complete space configuration with partial overwrite

        The values passed in ``test_space`` will override the default values
        in ``self.config``.

        Parameters
        ----------
        test_space: dic
            The configuration for the space.
        """
        space = copy.deepcopy(self.space)
        space.update(test_space)
        return space

    def create_space(self, space=None):
        """Create the space object

        Parameters
        ----------
        space: dict, optional
            Configuration of the search space. The default ``self.space`` will be used
            if ``space`` is ``None``.

        """
        return SpaceBuilder().build(space if space is not None else self.space)

    def observe_points(self, points, algo, objective=0):
        """Make the algorithm observe points

        Parameters
        ----------
        points: list of points
            Trials formatted as tuples of values
        algo: ``orion.algo.base.OptimizationAlgorithm``
            The algorithm used to observe points.
        objective: int, optional
            The base objective for the trials. All objectives
            will have value ``objective + i``. Defaults to 0.
        """
        algo.observe(
            points, [dict(objective=objective + i) for i in range(len(points))]
        )

    def get_num(self, num):
        """Force number of trials to suggest

        Some algorithms must be tested with specific number of suggests at a time (ex: ASHA).
        This method can be overriden to change ``num`` based on the special needs.
        """
        return num

    def force_observe(self, num, algo):
        """Force observe ``num`` points.

        Parameters
        ----------
        num: int
            Number of trials to suggest and observe.
        algo: ``orion.algo.base.OptimizationAlgorithm``
            The algorithm that must suggest and observe.

        Raises
        ------
        RuntimeError
            - If the algorithm returns duplicates. Algorithms may return duplicates across workers,
              but in sequential scenarios as here, it should not happen.
            - If the algorithm fails to sample any point at least 5 times.
        """
        objective = 0
        failed = 0
        MAX_FAILED = 5
        ids = set()
        while not algo.is_done and algo.n_observed < num and failed < MAX_FAILED:
            points = algo.suggest(self.get_num(num - algo.n_observed))
            if len(points) == 0:
                failed += 1
                continue
            for point in points:
                point_id = algo.get_id(point)
                if point_id in ids:
                    raise RuntimeError(f"algo suggested a duplicate: {point}")
                ids.add(point_id)
            ids |= set(algo.get_id(point) for point in points)
            self.observe_points(points, algo, objective)
            objective += len(points)

        if failed >= MAX_FAILED:
            raise RuntimeError(
                f"Algorithm cannot sample more than {algo.n_observed} points. Is it normal?"
            )

    def spy_phase(self, mocker, num, algo, attribute):
        """Force observe ``num`` points and then mock a given method to count calls.

        Parameters
        ----------
        mocker: ``pytest_mock.mocker``
            Mocker from ``pytest_mock``. Should be given by fixtures of the tests.
        num: int
            Number of trials to suggest and observe
        algo: ``orion.algo.base.OptimizationAlgorithm``
            The algorithm to test
        attribute: str
            The algorithm attribute or method to mock. The path is respective to the
            algorithm object. For example, a valid value would be `'space.sample'`
            which will mock ``algo.algorithm.space.sample``.
        """
        self.force_observe(num, algo)
        spy = spy_attr(mocker, algo.algorithm, attribute)
        return spy

    def assert_callbacks(self, spy, num, algo):
        """Callback to make special asserts at end of tests

        Override this method in algorithm test-suite to customize verifications done at end
        of tests.

        Parameters
        ----------
        spy: Mocked object
            Object mocked by ``BaseAlgoTests.spy_phase``.
        num: int
            number of points of the phase.
        algo: ``orion.algo.base.OptimizationAlgorithm``
            The algorithm being tested.
        """
        pass

    def assert_dim_type_supported(self, mocker, num, attr, test_space):
        """Test that a given dimension type is properly supported by the algorithm

        This will test that the algorithm sample points valid for the given type
        and that the algorithm can observe these points.

        Parameters
        ----------
        mocker: ``pytest_mock.mocker``
            Mocker from ``pytest_mock``. Should be given by fixtures of the tests.
        num: int
            Number of trials to suggest and observe
        algo: ``orion.algo.base.OptimizationAlgorithm``
            The algorithm to test
        attribute: str
            The algorithm attribute or method to mock. The path is respective to the
            algorithm object. For example, a valid value would be `'space.sample'`
            which will mock ``algo.algorithm.space.sample``.

        """
        space = self.create_space(self.update_space(test_space))
        algo = self.create_algo(space=space)

        spy = self.spy_phase(mocker, num, algo, attr)

        points = algo.suggest(1)
        assert points[0] in space
        spy.call_count == 1
        self.observe_points(points, algo, 1)
        self.assert_callbacks(spy, num, algo)

    def test_configuration(self):
        """Test that configuration property attribute contains all class arguments."""
        algo = self.create_algo()
        assert algo.configuration != self.create_algo(config={})
        assert algo.configuration == {self.algo_name: self.config}

    def test_get_id(self):
        """Test that the id hashing is valid"""
        space = self.create_space(
            space=self.update_space({"f": "fidelity(1, 10, base=2)"})
        )

        algo = self.create_algo(space=space)

        assert algo.get_id(["is here", 1]) == algo.get_id(["is here", 1])
        assert algo.get_id(["is here", 1]) != algo.get_id(["is here", 2])
        assert algo.get_id(["matters", 1]) != algo.get_id(["is here", 1])

        assert algo.get_id(["is here", 1], ignore_fidelity=False) == algo.get_id(
            ["is here", 1], ignore_fidelity=False
        )
        assert algo.get_id(["is here", 1], ignore_fidelity=False) != algo.get_id(
            ["is here", 2], ignore_fidelity=False
        )
        assert algo.get_id(["matters", 1], ignore_fidelity=False) != algo.get_id(
            ["is here", 1], ignore_fidelity=False
        )

        assert algo.get_id(["is here", 1], ignore_fidelity=True) == algo.get_id(
            ["is here", 1], ignore_fidelity=True
        )
        assert algo.get_id(["is here", 1], ignore_fidelity=True) != algo.get_id(
            ["is here", 2], ignore_fidelity=True
        )
        assert algo.get_id(["whatever", 1], ignore_fidelity=True) == algo.get_id(
            ["is here", 1], ignore_fidelity=True
        )

    @phase
    def test_seed_rng(self, mocker, num, attr):
        """Test that the seeding gives reproducibile results."""
        algo = self.create_algo()

        seed = numpy.random.randint(10000)
        algo.seed_rng(seed)
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(1)
        with pytest.raises(AssertionError):
            numpy.testing.assert_equal(points, algo.suggest(1))

        new_algo = self.create_algo()
        new_algo.seed_rng(seed)
        self.force_observe(algo.n_observed, new_algo)
        numpy.testing.assert_equal(points, new_algo.suggest(1))

        self.assert_callbacks(spy, num, new_algo)

    @phase
    def test_state_dict(self, mocker, num, attr):
        """Verify that resetting state makes sampling deterministic"""
        algo = self.create_algo()

        seed = numpy.random.randint(10000)
        algo.seed_rng(seed)
        spy = self.spy_phase(mocker, max(num, 1), algo, attr)
        state = algo.state_dict
        a = algo.suggest(1)[0]

        new_algo = self.create_algo()
        with pytest.raises(AssertionError):
            numpy.testing.assert_equal(a, new_algo.suggest(1)[0])

        new_algo.set_state(state)
        numpy.testing.assert_equal(a, new_algo.suggest(1)[0])

        self.assert_callbacks(spy, num, algo)

    @phase
    def test_suggest_n(self, mocker, num, attr):
        """Verify that suggest returns correct number of trials if ``num`` is specified in ``suggest``."""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(5)
        assert len(points) == 5

    @phase
    def test_has_suggested(self, mocker, num, attr):
        """Verify that algorithm detects correctly if a point was suggested"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        a = algo.suggest(1)[0]
        assert algo.has_suggested(a)
        # NOTE: not algo.has_suggested(some random point) is tested in test_has_suggested_statedict

    @phase
    def test_has_suggested_statedict(self, mocker, num, attr):
        """Verify that algorithm detects correctly if a point was suggested even when state was restored."""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)

        a = algo.suggest(1)[0]
        state = algo.state_dict
        assert algo.has_suggested(a)

        algo = self.create_algo()
        assert not algo.has_suggested(a)

        algo.set_state(state)
        assert algo.has_suggested(a)

    @phase
    def test_observe(self, mocker, num, attr):
        """Verify that algorithm observes point without any issues"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)

        a = algo.space.sample()[0]
        algo.observe([a], [dict(objective=1)])

        b = algo.suggest(1)[0]
        algo.observe([b], [dict(objective=2)])

    @phase
    def test_has_observed(self, mocker, num, attr):
        """Verify that algorithm detects correctly if a point was observed"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)

        a = algo.suggest(1)[0]
        assert not algo.has_observed(a)
        algo.observe([a], [dict(objective=1)])
        assert algo.has_observed(a)

        b = algo.suggest(1)[0]
        assert not algo.has_observed(b)
        algo.observe([b], [dict(objective=2)])
        assert algo.has_observed(b)

    @phase
    def test_has_observed_statedict(self, mocker, num, attr):
        """Verify that algorithm detects correctly if a point was observed even when state was restored."""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)

        a = algo.suggest(1)[0]
        algo.observe([a], [dict(objective=1)])
        state = algo.state_dict

        algo = self.create_algo()
        assert not algo.has_observed(a)
        algo.set_state(state)
        assert algo.has_observed(a)

        b = algo.suggest(1)[0]
        algo.observe([b], [dict(objective=2)])
        state = algo.state_dict

        algo = self.create_algo()
        assert not algo.has_observed(b)
        algo.set_state(state)
        assert algo.has_observed(b)

    @phase
    def test_n_suggested(self, mocker, num, attr):
        """Verify that algorithm returns correct number of suggested points"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        assert algo.n_suggested == num
        algo.suggest(1)
        assert algo.n_suggested == num + 1

    @phase
    def test_n_observed(self, mocker, num, attr):
        """Verify that algorithm returns correct number of observed points"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        assert algo.n_observed == num
        points = algo.suggest(1)
        assert algo.n_observed == num
        self.observe_points(points, algo)
        assert algo.n_observed == num + 1

    @phase
    def test_real_data(self, mocker, num, attr):
        """Test that algorithm supports real dimesions"""
        self.assert_dim_type_supported(mocker, num, attr, {"x": "uniform(0, 5)"})

    @phase
    def test_int_data(self, mocker, num, attr):
        """Test that algorithm supports integer dimesions"""
        self.assert_dim_type_supported(
            mocker, num, attr, {"x": "uniform(0, 5000, discrete=True)"}
        )

    @phase
    def test_cat_data(self, mocker, num, attr):
        """Test that algorithm supports categorical dimesions"""
        self.assert_dim_type_supported(
            mocker,
            num,
            attr,
            {  # Add 3 dims so that there exists many possible trials for the test
                "x": "choices(['a', 0.2, 1, None])",
                "y": "choices(['a', 0.2, 1, None])",
                "z": "choices(['a', 0.2, 1, None])",
            },
        )

    @phase
    def test_logreal_data(self, mocker, num, attr):
        """Test that algorithm supports logreal dimesions"""
        self.assert_dim_type_supported(mocker, num, attr, {"x": "loguniform(1, 5)"})

    @phase
    def test_logint_data(self, mocker, num, attr):
        """Test that algorithm supports loginteger dimesions"""
        self.assert_dim_type_supported(
            mocker, num, attr, {"x": "loguniform(1, 100, discrete=True)"}
        )

    @phase
    def test_shape_data(self, mocker, num, attr):
        """Test that algorithm supports dimesions with shape"""
        self.assert_dim_type_supported(
            mocker, num, attr, {"x": "uniform(0, 5, shape=(3, 2))"}
        )

    def test_is_done_cardinality(self):
        """Test that algorithm will stop when cardinality is reached"""
        space = self.update_space(
            {
                "x": "uniform(0, 4, discrete=True)",
                "y": "choices(['a', 'b', 'c'])",
                "z": "loguniform(1, 6, discrete=True)",
            }
        )
        space = self.create_space(space)
        assert space.cardinality == 5 * 3 * 6

        algo = self.create_algo(space=space)
        for i, (x, y, z) in enumerate(itertools.product(range(5), "abc", range(1, 7))):
            assert not algo.is_done
            n = algo.n_suggested
            algo.observe([[x, y, z]], [dict(objective=i)])
            assert algo.n_suggested == n + 1

        assert i + 1 == space.cardinality

        assert algo.is_done

    def test_is_done_max_trials(self):
        """Test that algorithm will stop when max trials is reached"""
        algo = self.create_algo()
        self.force_observe(self.max_trials, algo)

        assert algo.is_done

    def test_optimize_branin(self):
        """Test that algorithm optimizes somehow (this is on-par with random search)"""
        MAX_TRIALS = 20
        task = Branin()
        space = self.create_space(task.get_search_space())
        algo = self.create_algo(config={}, space=space)
        algo.algorithm.max_trials = MAX_TRIALS
        safe_guard = 0
        points = []
        objectives = []
        while points or not algo.is_done:
            if safe_guard >= MAX_TRIALS:
                break

            if not points:
                points = algo.suggest(MAX_TRIALS - len(objectives))

            point = points.pop(0)
            results = task(*point)
            objectives.append(results[0]["value"])
            algo.observe([point], [dict(objective=objectives[-1])])
            safe_guard += 1

        assert algo.is_done
        assert min(objectives) <= 10



"""Collection of tests for :mod:`orion.plotting.orion.plotting.PlotAccessor`."""
import pytest

from orion.core.worker.experiment import Experiment
from orion.plotting.base import PlotAccessor
from orion.testing import create_experiment

SINGLE_EXPERIMENT_PLOTS = (
    "lpi",
    "regret",
    "parallel_coordinates",
    "partial_dependencies",
)

config = dict(
    name="experiment-name",
    space={"x": "uniform(0, 200)"},
    metadata={
        "user": "test-user",
        "orion_version": "XYZ",
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir="",
    algorithms={"random": {"seed": 1}},
    producer={"strategy": "NoParallelStrategy"},
)


trial_config = {
    "experiment": 0,
    "status": "new",  # new, reserved, suspended, completed, broken
    "worker": None,
    "start_time": None,
    "end_time": None,
    "heartbeat": None,
    "results": [],
    "params": [],
}


def check_plot(plot, kind):
    """Verifies that existence of the plot"""
    assert plot
    assert kind.replace("_", " ") in plot.layout.title.text.lower()


def test_init_require_experiment():
    """Tests that a `PlotAccessor` requires an instance of `ExperimentClient`"""
    with pytest.raises(ValueError) as exception:
        PlotAccessor(None)

    assert "Parameter 'experiment' is None" in str(exception.value)


def test_call_nonexistent_kind():
    """Tests that specifying a non existent kind will fail"""
    with create_experiment(config, trial_config, ["completed"]) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        with pytest.raises(ValueError) as exception:
            pa(kind="nonexistent")

        assert "Plot of kind 'nonexistent' is not one of" in str(exception.value)


def test_regret_is_default_plot():
    """Tests that the regret plot is the default plot"""
    with create_experiment(config, trial_config, ["completed"]) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = pa()

        check_plot(plot, "regret")


@pytest.mark.parametrize("kind", SINGLE_EXPERIMENT_PLOTS)
def test_kind(kind):
    """Tests that a plot can be created from specifying `kind` as a parameter."""
    with create_experiment(config, trial_config, ["completed"]) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = pa(kind=kind)

        check_plot(plot, kind)


@pytest.mark.parametrize("kind", SINGLE_EXPERIMENT_PLOTS)
def test_call_to(kind):
    """Tests instance calls to `PlotAccessor.{kind}()`"""
    with create_experiment(config, trial_config, ["completed"]) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = getattr(pa, kind)()

        check_plot(plot, kind)


@pytest.mark.parametrize("kind", SINGLE_EXPERIMENT_PLOTS)
def test_emtpy(kind):
    """Tests instance calls to `PlotAccessor.{kind}()`"""
    with create_experiment(config, trial_config, []) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        getattr(pa, kind)()


@pytest.mark.parametrize("kind", SINGLE_EXPERIMENT_PLOTS)
def test_with_evc_tree(monkeypatch, kind):
    """Tests that the plotly backend returns a plotly object"""

    WITH_EVC_TREE = True

    original_to_pandas = Experiment.to_pandas

    def mock_to_pandas(self, with_evc_tree):
        assert with_evc_tree is WITH_EVC_TREE
        return original_to_pandas(self, with_evc_tree)

    monkeypatch.setattr(
        "orion.core.worker.experiment.Experiment.to_pandas", mock_to_pandas
    )

    with create_experiment(config, trial_config, ["completed"]) as (_, _, experiment):
        pa = PlotAccessor(experiment)

        WITH_EVC_TREE = True
        plot = getattr(pa, kind)(with_evc_tree=WITH_EVC_TREE)
        check_plot(plot, kind)

        WITH_EVC_TREE = False
        plot = getattr(pa, kind)(with_evc_tree=WITH_EVC_TREE)
        check_plot(plot, kind)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module running the optimization command
=======================================

Gets an experiment and iterates over it until one of the exit conditions is met

"""

import logging
import signal

import orion.core
import orion.core.io.experiment_builder as experiment_builder
from orion.client.experiment import ExperimentClient
from orion.core.cli import base as cli
from orion.core.cli import evc as evc_cli
from orion.core.utils.exceptions import (
    BrokenExperiment,
    InexecutableUserScript,
    MissingResultFile,
)
from orion.core.utils.format_terminal import format_stats
from orion.core.worker.consumer import Consumer
from orion.core.worker.producer import Producer

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = "Conducts hyperparameter optimization"
DESCRIPTION = """
This command starts hyperparameter optimization process for the user-provided model using the
configured optimization algorithm and search space.
"""


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    hunt_parser = parser.add_parser(
        "hunt", help=SHORT_DESCRIPTION, description=DESCRIPTION
    )

    orion_group = cli.get_basic_args_group(
        hunt_parser, group_name="Hunt arguments", group_help=""
    )

    orion.core.config.experiment.add_arguments(
        orion_group,
        rename=dict(max_broken="--exp-max-broken", max_trials="--exp-max-trials"),
    )

    orion_group.add_argument(
        "--max-trials",
        type=int,
        metavar="#",
        help="(DEPRECATED) This argument will be removed in v0.3. Use --exp-max-trials instead",
    )

    orion_group.add_argument(
        "--init-only",
        default=False,
        action="store_true",
        help="Only create the experiment and register in database, but do not execute any trial.",
    )

    worker_args_group = hunt_parser.add_argument_group(
        "Worker arguments (optional)",
        description="Arguments to automatically resolved branching events.",
    )

    orion.core.config.worker.add_arguments(
        worker_args_group,
        rename=dict(max_broken="--worker-max-broken", max_trials="--worker-max-trials"),
    )

    evc_cli.get_branching_args_group(hunt_parser)

    cli.get_user_args_group(hunt_parser)

    hunt_parser.set_defaults(func=main)
    hunt_parser.set_defaults(help_empty=True)  # Print help if command is empty

    return hunt_parser


COMPLETION_MESSAGE = """\
Hints
=====

Info
----

To get more information on the experiment, run the command

orion info --name {experiment.name} --version {experiment.version}

"""


NONCOMPLETED_MESSAGE = """\
Status
------

To get the status of the trials, run the command

orion status --name {experiment.name} --version {experiment.version}


For a detailed view with status of each trial listed, use the argument `--all`

orion status --name {experiment.name} --version {experiment.version} --all

"""

# pylint: disable = unused-argument
def _handler(signum, frame):
    log.error("Oríon has been interrupted.")
    raise KeyboardInterrupt


# pylint:disable=unused-argument
def on_error(client, trial, error, worker_broken_trials):
    """If the script is not executable, don't waste time and raise right away"""
    if isinstance(error, (InexecutableUserScript, MissingResultFile)):
        raise error

    return True


# pylint:disable=too-many-arguments
def workon(
    experiment,
    n_workers=None,
    max_trials=None,
    max_broken=None,
    max_idle_time=None,
    heartbeat=None,
    user_script_config=None,
    interrupt_signal_code=None,
    ignore_code_changes=None,
    executor=None,
    executor_configuration=None,
):
    """Try to find solution to the search problem defined in `experiment`."""
    producer = Producer(experiment, max_idle_time)
    consumer = Consumer(
        experiment,
        user_script_config,
        interrupt_signal_code,
        ignore_code_changes,
    )

    client = ExperimentClient(experiment, producer, heartbeat=heartbeat)

    if executor is None:
        executor = orion.core.config.worker.executor

    if executor_configuration is None:
        executor_configuration = orion.core.config.worker.executor_configuration

    log.debug("Starting workers")
    with client.tmp_executor(executor, n_workers=n_workers, **executor_configuration):
        try:
            client.workon(
                consumer,
                n_workers=n_workers,
                max_trials_per_worker=max_trials,
                max_broken=max_broken,
                trial_arg="trial",
                on_error=on_error,
            )
        except BrokenExperiment as e:
            print(e)

    if client.is_done:
        print("Search finished successfully")

    print("\n" + format_stats(client))

    print("\n" + COMPLETION_MESSAGE.format(experiment=client))
    if not experiment.is_done:
        print(NONCOMPLETED_MESSAGE.format(experiment=client))


def main(args):
    """Build experiment and execute hunt command"""
    args["root"] = None
    args["leafs"] = []
    # TODO: simplify when parameter parsing is refactored
    experiment = experiment_builder.build_from_args(args)

    if args["init_only"]:
        return

    config = experiment_builder.get_cmd_config(args)
    worker_config = orion.core.config.worker.to_dict()
    if config.get("worker"):
        worker_config.update(config.get("worker"))

    signal.signal(signal.SIGTERM, _handler)

    workon(
        experiment,
        ignore_code_changes=config["branching"].get("ignore_code_changes"),
        **worker_config
    )



import joblib

from orion.executor.base import BaseExecutor


class Joblib(BaseExecutor):
    def __init__(self, n_workers=-1, backend="loky", **config):
        super(Joblib, self).__init__(n_workers=n_workers)
        self.backend = backend
        self.config = config

        self.joblib_parallel = joblib.parallel_backend(
            self.backend, n_jobs=self.n_workers, **self.config
        )

    def __getstate__(self):
        state = super(Joblib, self).__getstate__()
        state["backend"] = self.backend
        state["config"] = self.config
        return state

    def __setstate__(self, state):
        super(Joblib, self).__setstate__(state)
        self.backend = state["backend"]
        self.config = state["config"]

        self.joblib_parallel = joblib.parallel_backend(
            self.backend, n_jobs=self.n_workers, **self.config
        )

    def wait(self, futures):
        return joblib.Parallel(n_jobs=self.n_workers)(futures)

    def submit(self, function, *args, **kwargs):
        return joblib.delayed(function)(*args, **kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        self.joblib_parallel.unregister()
        super(Joblib, self).__exit__(exc_type, exc_value, traceback)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple one dimensional example with noise level for a possible user's script."""
import argparse
import random

from orion.client import report_results


def function(x, noise):
    """Evaluate partial information of a quadratic."""
    z = (x - 34.56789) * random.gauss(0, noise)
    return 4 * z ** 2 + 23.4, 8 * z


def execute():
    """Execute a simple pipeline as an example."""
    # 1. Receive inputs as you want
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", type=float, required=True)
    parser.add_argument("--fidelity", type=int, default=10)
    inputs = parser.parse_args()

    assert 0 <= inputs.fidelity <= 10

    noise = (1 - inputs.fidelity / 10) + 0.0001

    # 2. Perform computations
    y, dy = function(inputs.x, noise)

    # 3. Gather and report results
    results = list()
    results.append(dict(name="example_objective", type="objective", value=y))
    results.append(dict(name="example_gradient", type="gradient", value=[dy]))

    report_results(results)


if __name__ == "__main__":
    execute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the default commands for demo purposes."""
import os

import pytest

import orion.core.cli
from orion.storage.base import get_storage


def get_user_corneau():
    """Return user corneau (to mock getpass.getuser)"""
    return "corneau"


def test_insert_invalid_experiment(storage, monkeypatch, capsys):
    """Test the insertion of an invalid experiment"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    returncode = orion.core.cli.main(
        [
            "insert",
            "-n",
            "dumb_experiment",
            "-c",
            "./orion_config_random.yaml",
            "./black_box.py",
            "-x=1",
        ]
    )

    assert returncode == 1

    captured = capsys.readouterr().err

    assert "Error: No experiment with given name 'dumb_experiment'"


@pytest.mark.usefixtures("only_experiments_db", "version_XYZ")
def test_insert_single_trial(storage, monkeypatch, script_path):
    """Try to insert a single trial"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    orion.core.cli.main(
        [
            "insert",
            "-n",
            "test_insert_normal",
            "-c",
            "./orion_config_random.yaml",
            "blabla",
            "-x=1",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "test_insert_normal"}))

    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp

    trials = list(storage.fetch_trials(uid=exp["_id"]))

    assert len(trials) == 1

    trial = trials[0]

    assert trial.status == "new"
    assert trial.params["/x"] == 1


@pytest.mark.usefixtures("only_experiments_db", "version_XYZ")
def test_insert_single_trial_default_value(storage, monkeypatch):
    """Try to insert a single trial using a default value"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    orion.core.cli.main(
        [
            "insert",
            "-n",
            "test_insert_normal",
            "-c",
            "./orion_config_random.yaml",
            "./black_box.py",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "test_insert_normal"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp

    trials = list(storage.fetch_trials(uid=exp["_id"]))

    assert len(trials) == 1

    trial = trials[0]

    assert trial.status == "new"
    assert trial.params["/x"] == 1


@pytest.mark.usefixtures("only_experiments_db")
def test_insert_with_no_default_value(monkeypatch):
    """Try to insert a single trial by omitting a namespace with no default value"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(
            [
                "insert",
                "-n",
                "test_insert_missing_default_value",
                "-c",
                "./orion_config_random.yaml",
                "./black_box.py",
            ]
        )

    assert "Dimension /x is unspecified and has no default value" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db")
def test_insert_with_incorrect_namespace(monkeypatch):
    """Try to insert a single trial with a namespace not inside the experiment space"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(
            [
                "insert",
                "-n",
                "test_insert_normal",
                "-c",
                "./orion_config_random.yaml",
                "./black_box.py",
                "-p=4",
            ]
        )

    assert "Found namespace outside of experiment space : /p" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db")
def test_insert_with_outside_bound_value(monkeypatch):
    """Try to insert a single trial with value outside the distribution's interval"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(
            [
                "insert",
                "-n",
                "test_insert_two_hyperparameters",
                "-c",
                "./orion_config_random.yaml",
                "./black_box.py",
                "-x=4",
                "-y=100",
            ]
        )
    assert "Value 100 is outside of" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db", "version_XYZ")
def test_insert_two_hyperparameters(storage, monkeypatch):
    """Try to insert a single trial with two hyperparameters"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)
    orion.core.cli.main(
        [
            "insert",
            "-n",
            "test_insert_two_hyperparameters",
            "-c",
            "./orion_config_random.yaml",
            "./black_box.py",
            "-x=1",
            "-y=2",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "test_insert_two_hyperparameters"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp

    trials = list(storage.fetch_trials(uid=exp["_id"]))

    assert len(trials) == 1

    trial = trials[0]

    assert trial.status == "new"
    assert trial.params["/x"] == 1
    assert trial.params["/y"] == 2


def test_insert_with_version(storage, monkeypatch, script_path):
    """Try to insert a single trial inside a specific version"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "experiment",
            "-c",
            "./orion_config_random.yaml",
            script_path,
            "-x~normal(0,1)",
        ]
    )
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "-n",
            "experiment",
            "-c",
            "./orion_config_random.yaml",
            script_path,
            "-x~normal(0,1)",
            "-y~+normal(0,1)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "experiment", "version": 1}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp

    trials = list(storage.fetch_trials(uid=exp["_id"]))
    assert len(trials) == 0

    orion.core.cli.main(
        [
            "insert",
            "-n",
            "experiment",
            "--version",
            "1",
            "-c",
            "./orion_config_random.yaml",
            script_path,
            "-x=1",
        ]
    )

    trials = list(storage.fetch_trials(uid=exp["_id"]))

    assert len(trials) == 1


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main(["insert"])

    captured = capsys.readouterr().out

    assert "usage:" in captured
    assert "Traceback" not in captured


def test_no_name(capsys):
    """Try to run the command without providing an experiment name"""
    returncode = orion.core.cli.main(["insert", "--version", "1"])
    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured == "Error: No name provided for the experiment.\n"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Task for RosenBrock Function
==============================
"""
import numpy

from orion.benchmark.task.base import BaseTask


class RosenBrock(BaseTask):
    """`RosenBrock function <http://infinity77.net/global_optimization/test_functions_nd_R.html#go_benchmark.Rosenbrock>`_
    as benchmark task"""

    def __init__(self, max_trials=20, dim=2):
        self.dim = dim
        super(RosenBrock, self).__init__(max_trials=max_trials, dim=dim)

    def call(self, x):
        """Evaluate a n-D rosenbrock function."""
        x = numpy.asarray(x)
        summands = 100 * (x[1:] - x[:-1] ** 2) ** 2 + (1 - x[:-1]) ** 2
        y = numpy.sum(summands)
        return [dict(name="rosenbrock", type="objective", value=y)]

    def get_search_space(self):
        """Return the search space for the task objective function"""
        rspace = {"x": "uniform(-5, 10, shape={})".format(self.dim)}

        return rspace



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Average Rank Assessment
========================
"""

from collections import defaultdict

from orion.benchmark.assessment.base import BaseAssess
from orion.plotting.base import rankings


class AverageRank(BaseAssess):
    """
    Evaluate the average performance (objective value) between different search algorithms from
    the rank perspective at different time steps (trial number).
    The performance (objective value) used for a trial will the best result until the trial.
    """

    def __init__(self, task_num=1):
        super(AverageRank, self).__init__(task_num=task_num)

    def analysis(self, task, experiments):
        """
        Generate a `plotly.graph_objects.Figure` to display average rankings
        between different search algorithms.

        task: str
            Name of the task
        experiments: list
            A list of (task_index, experiment), where task_index is the index of task to run for
            this assessment, and experiment is an instance of `orion.core.worker.experiment`.
        """
        algorithm_groups = defaultdict(list)

        for _, exp in experiments:
            algorithm_name = list(exp.configuration["algorithms"].keys())[0]
            algorithm_groups[algorithm_name].append(exp)

        return rankings(algorithm_groups)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base definition of Task
========================
"""

from abc import ABC, abstractmethod

from orion.core.utils import Factory


class BaseTask(ABC):
    """Base class describing what a task can do.
    A task will define the objective function and search space of it.

    Parameters
    ----------
    max_trials : int
       Max number of trials the experiment will run against this task.
    kwargs : dict
       Configurable parameters of the task, a particular task
       implementation can have its own parameters.
    """

    def __init__(self, max_trials, **kwargs):
        self.trials_num = max_trials
        self._param_names = kwargs
        self._param_names["max_trials"] = max_trials

    @abstractmethod
    def call(self, *args, **kwargs):
        """
        Define the black box function to optimize, the function will expect hyper-parameters to
        search and return objective values of trial with the hyper-parameters.
        """
        pass

    def __call__(self, *args, **kwargs):
        """
        All tasks will be callable by default,
        and method `call()` will be executed when a task is called directly.
        """
        return self.call(*args, **kwargs)

    @property
    def max_trials(self):
        """Return the max number of trials to run for the"""
        return self.trials_num

    @abstractmethod
    def get_search_space(self):
        """Return the search space for the task objective function"""
        pass

    @property
    def configuration(self):
        """Return the configuration of the task."""
        return {self.__class__.__qualname__: self._param_names}


# pylint: disable=too-few-public-methods,abstract-method
class BenchmarkTask(BaseTask, metaclass=Factory):
    """Class used to inject dependency on an task implementation."""

    pass



"""Perform tests for the REST endpoint `/trials`"""
import copy
import datetime
import random

import falcon

from orion.core.io import experiment_builder
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage

base_experiment = dict(
    name="experiment-name",
    space={"x": "uniform(0, 200)"},
    metadata={
        "user": "test-user",
        "orion_version": "x.y.z",
        "datetime": datetime.datetime(1, 1, 1),
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    refers={},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir="",
    algorithms={"random": {"seed": 1}},
)

base_trial = {
    "experiment": None,
    "status": "new",
    "worker": None,
    "submit_time": datetime.datetime(1, 1, 1),
    "start_time": datetime.datetime(1, 1, 1, second=10),
    "end_time": datetime.datetime(1, 1, 2),
    "heartbeat": None,
    "results": [
        {"name": "loss", "type": "objective", "value": 0.05},
        {"name": "a", "type": "statistic", "value": 10},
        {"name": "b", "type": "statistic", "value": 5},
    ],
    "params": [
        {"name": "x", "type": "real", "value": 10.0},
    ],
}


def add_experiment(**kwargs):
    """Adds experiment to the dummy orion instance"""
    base_experiment.update(copy.deepcopy(kwargs))
    experiment_builder.build(
        branching=dict(branch_from=base_experiment["name"]), **base_experiment
    )


def add_trial(experiment: int, status: str = None, **kwargs):
    """
    Add trials to the dummy orion instance

    Parameters
    ----------
    experiment
        The ID of the experiment (stored in Experiment._id)
    status
        The status of the trial to add.
    **kwargs
        Any other attribute to modify from the base trial config.
    """
    if not status:
        status = random.choice(Trial.allowed_stati)

    kwargs["experiment"] = experiment
    kwargs["status"] = status

    base_trial.update(copy.deepcopy(kwargs))
    get_storage().register_trial(Trial(**base_trial))


def test_root_endpoint_not_supported(client):
    """Tests that the server return a 404 when accessing trials/ with no experiment"""
    response = client.simulate_get("/trials")

    assert response.status == "404 Not Found"
    if falcon.__version__ < "3.0.0":
        assert not response.json
    else:
        assert response.json == {"title": "404 Not Found"}


class TestTrialCollection:
    """Tests trials/:experiment_name"""

    def test_trials_for_unknown_experiment(self, client):
        """Tests that an unknown experiment returns a not found error"""
        response = client.simulate_get("/trials/unknown-experiment")

        assert response.status == "404 Not Found"
        assert response.json == {
            "title": "Experiment not found",
            "description": 'Experiment "unknown-experiment" does not exist',
        }

    def test_unknown_parameter(self, client):
        """
        Tests that if an unknown parameter is specified in
        the query string, an error is returned even if the experiment doesn't exist.
        """
        expected_error_message = (
            'Parameter "unknown" is not supported. '
            "Expected one of ['ancestors', 'status', 'version']."
        )

        response = client.simulate_get("/trials/a?unknown=true")

        assert response.status == "400 Bad Request"
        assert response.json == {
            "title": "Invalid parameter",
            "description": expected_error_message,
        }

        add_experiment(name="a", version=1, _id=1)

        response = client.simulate_get("/trials/a?unknown=true")

        assert response.status == "400 Bad Request"
        assert response.json == {
            "title": "Invalid parameter",
            "description": expected_error_message,
        }

    def test_trials_for_latest_version(self, client):
        """Tests that it returns the trials of the latest version of the experiment"""
        add_experiment(name="a", version=1, _id=1)
        add_experiment(name="a", version=2, _id=2)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=1, id_override="02")
        add_trial(experiment=2, id_override="03")

        response = client.simulate_get("/trials/a")

        assert response.status == "200 OK"
        assert response.json == [{"id": "01"}, {"id": "03"}]

    def test_trials_for_specific_version(self, client):
        """Tests specific version of experiment"""
        add_experiment(name="a", version=1, _id=1)
        add_experiment(name="a", version=2, _id=2)
        add_experiment(name="a", version=3, _id=3)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=3, id_override="02")

        # Happy case
        response = client.simulate_get("/trials/a?version=2")

        assert response.status == "200 OK"
        assert response.json == [{"id": "01"}]

        # Version doesn't exist
        response = client.simulate_get("/trials/a?version=4")

        assert response.status == "404 Not Found"
        assert response.json == {
            "title": "Experiment not found",
            "description": 'Experiment "a" has no version "4"',
        }

    def test_trials_for_all_versions(self, client):
        """Tests that trials from all ancestors are shown"""
        add_experiment(name="a", version=1, _id=1)
        add_experiment(name="a", version=2, _id=2)
        add_experiment(name="a", version=3, _id=3)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=3, id_override="02")

        # Happy case default
        response = client.simulate_get("/trials/a?ancestors=true")

        assert response.status == "200 OK"
        assert response.json == [{"id": "00"}, {"id": "01"}, {"id": "02"}]

        # Happy case explicitly false (call latest version test)
        response = client.simulate_get("/trials/a?ancestors=false")

        assert response.status == "200 OK"
        assert response.json == [{"id": "02"}]

        # Not a boolean parameter
        response = client.simulate_get("/trials/a?ancestors=42")

        assert response.status == "400 Bad Request"
        assert response.json == {
            "title": "Invalid parameter",
            "description": 'The "ancestors" parameter is invalid. '
            'The value of the parameter must be "true" or "false".',
        }

    def test_trials_by_status(self, client):
        """Tests that trials are returned"""
        add_experiment(name="a", version=1, _id=1)

        # There exist no trial of the given status in an empty experiment
        response = client.simulate_get("/trials/a?status=completed")

        assert response.status == "200 OK"
        assert response.json == []

        # There exist no trial of the given status while other status are present
        add_trial(experiment=1, id_override="00", status="broken")

        response = client.simulate_get("/trials/a?status=completed")
        assert response.status == "200 OK"
        assert response.json == []

        # There exist at least one trial of the given status
        add_trial(experiment=1, id_override="01", status="completed")

        response = client.simulate_get("/trials/a?status=completed")

        assert response.status == "200 OK"
        assert response.json == [{"id": "01"}]

        # Status doesn't exist
        response = client.simulate_get("/trials/a?status=invalid")

        assert response.status == "400 Bad Request"
        error_message = (
            "Invalid status value. "
            "Expected one of "
            "['new', 'reserved', 'suspended', 'completed', 'interrupted', 'broken']"
        )
        assert response.json == {
            "title": "Invalid parameter",
            "description": 'The "status" parameter is invalid. '
            "The value of the parameter must be one of "
            "['new', 'reserved', 'suspended', 'completed', "
            "'interrupted', 'broken']",
        }

    def test_trials_by_from_specific_version_by_status_with_ancestors(self, client):
        """Tests that mixing parameters work as intended"""
        add_experiment(name="a", version=1, _id=1)
        add_experiment(name="b", version=1, _id=2)
        add_experiment(name="a", version=2, _id=3)
        add_experiment(name="a", version=3, _id=4)

        add_trial(experiment=1, id_override="00", status="completed")
        add_trial(experiment=3, id_override="01", status="broken")
        add_trial(experiment=3, id_override="02", status="completed")
        add_trial(experiment=2, id_override="03", status="completed")

        response = client.simulate_get(
            "/trials/a?ancestors=true&version=2&status=completed"
        )

        assert response.status == "200 OK"
        assert response.json == [{"id": "00"}, {"id": "02"}]


class TestTrialItem:
    """Tests trials/:experiment_name/:trial_id"""

    def test_unknown_experiment(self, client):
        """Tests that an unknown experiment returns a not found error"""
        response = client.simulate_get("/trials/unknown-experiment/a-trial")

        assert response.status == "404 Not Found"
        assert response.json == {
            "title": "Experiment not found",
            "description": 'Experiment "unknown-experiment" does not exist',
        }

    def test_unknown_trial(self, client):
        """Tests that an unknown experiment returns a not found error"""
        add_experiment(name="a", version=1, _id=1)

        response = client.simulate_get("/trials/a/unknown-trial")

        assert response.status == "404 Not Found"
        assert response.json == {
            "title": "Trial not found",
            "description": 'Trial "unknown-trial" does not exist',
        }

    def test_get_trial(self, client):
        """Tests that an existing trial is returned according to the API specification"""
        add_experiment(name="a", version=1, _id=1)
        add_trial(experiment=1, id_override="00", status="completed")
        add_trial(experiment=1, id_override="01", status="completed")

        response = client.simulate_get("/trials/a/01")

        assert response.status == "200 OK"
        assert response.json == {
            "id": "01",
            "submitTime": "0001-01-01 00:00:00",
            "startTime": "0001-01-01 00:00:10",
            "endTime": "0001-01-02 00:00:00",
            "parameters": {"x": 10.0},
            "objective": 0.05,
            "statistics": {"a": 10, "b": 5},
        }
